@@ -18,7 +18,7 @@
 /***************************************************************************
  * Description: Experimental bi-directionl protocol handler.               *
  * Author:      Gal Shachor <shachor@il.ibm.com>                           *
- * Version:     $Revision: 751213 $                                           *
+ * Version:     $Revision: 1650098 $                                           *
  ***************************************************************************/
 #ifndef JK_AJP13_H
 #define JK_AJP13_H
@@ -36,7 +36,9 @@ extern "C"
 
 #define AJP13_DEF_HOST              ("localhost")
 #define AJP13_DEF_PORT              (8009)
-#define AJP13_READ_BUF_SIZE         (8*1024)
+#define AJP13_DEF_PACKET_SIZE       (8*1024)
+#define AJP13_MAX_PACKET_SIZE       (64*1024)
+#define AJP13_PACKET_SIZE_ALIGN     (1024)
 #define AJP13_DEF_CACHE_SZ          (1)
 #define JK_INTERNAL_ERROR           (-2)
 #define JK_FATAL_ERROR              (-3)
@@ -48,8 +50,8 @@ extern "C"
 #define JK_STATUS_FATAL_ERROR       (-9)
 #define JK_REPLY_TIMEOUT            (-10)
 #define JK_AJP_PROTOCOL_ERROR       (-11)
+#define JK_BUSY_ERROR               (-12)
 
-#define AJP13_MAX_SEND_BODY_SZ      (DEF_BUFFER_SZ - 6)
 #define AJP13_DEF_TIMEOUT           (0) /* Idle timout for pooled connections */
 
 /*
