@@ -19,7 +19,7 @@
  * Description: common stuff for bi-directional protocol ajp13/ajp14.      *
  * Author:      Gal Shachor <shachor@il.ibm.com>                           *
  * Author:      Henri Gomez <hgomez@apache.org>                            *
- * Version:     $Revision: 1199891 $                                          *
+ * Version:     $Revision: 1665454 $                                          *
  ***************************************************************************/
 
 #ifndef JK_AJP_COMMON_H
@@ -47,10 +47,6 @@ extern "C"
 #define JK_AJP_STATE_TEXT_MAX           (JK_AJP_STATE_PROBE)
 #define JK_AJP_STATE_TEXT_DEF           (JK_AJP_STATE_TEXT_IDLE)
 
-/* We accept doing global maintenance if we are */
-/* JK_AJP_MAINTAIN_TOLERANCE seconds early. */
-#define JK_AJP_MAINTAIN_TOLERANCE (2)
-
 /*
  * Conditional request attributes
  *
@@ -72,11 +68,22 @@ extern "C"
 
 /*
  * AJP private request attributes
- *
+ */
+/*
+ * The following request attribute is recognized by Tomcat
+ * to contain the name of the ssl protocol used
+ */
+#define SC_A_SSL_PROTOCOL       ("AJP_SSL_PROTOCOL")
+/*
  * The following request attribute is recognized by Tomcat
  * to contain the forwarded remote port.
  */
 #define SC_A_REQ_REMOTE_PORT    ("AJP_REMOTE_PORT")
+/*
+ * The following request attribute is recognized by Tomcat
+ * to contain the forwarded local ip address.
+ */
+#define SC_A_REQ_LOCAL_ADDR     ("AJP_LOCAL_ADDR")
 
 /*
  * JK public request attributes
@@ -236,6 +243,12 @@ extern "C"
 #define AJP_CPING_CONNECT         (1)   /* Send cping on fresh connection */
 #define AJP_CPING_PREPOST         (2)   /* Send cping before sending request */
 #define AJP_CPING_INTERVAL        (4)   /* Send cping on regular intervals */
+#define AJP_CPING_MAX             (AJP_CPING_INTERVAL)
+
+#define AJP_CPING_CONNECT_TEXT    ('C')   /* Send cping on fresh connection */
+#define AJP_CPING_PREPOST_TEXT    ('P')   /* Send cping before sending request */
+#define AJP_CPING_INTERVAL_TEXT   ('I')   /* Send cping on regular intervals */
+#define AJP_CPING_ALL_TEXT        ('A')   /* Send cping on regular intervals */
 
 
 #define RECOVER_ABORT_IF_TCGETREQUEST    0x0001 /* DON'T RECOVER IF TOMCAT FAILS AFTER RECEIVING REQUEST */
@@ -244,8 +257,6 @@ extern "C"
 #define RECOVER_ALWAYS_HTTP_HEAD         0x0008 /* RECOVER HTTP HEAD REQUESTS, EVEN IF ABORT OPTIONS ARE SET */
 #define RECOVER_ALWAYS_HTTP_GET          0x0010 /* RECOVER HTTP GET REQUESTS, EVEN IF ABORT OPTIONS ARE SET */
 
-#define JK_MAX_HTTP_STATUS_FAILS   32   /* Should be enough for most 400 and 500 statuses */
-
 struct jk_res_data
 {
     int status;
@@ -277,7 +288,7 @@ struct ajp_worker
     /* Shared memory worker data */
     jk_shm_ajp_worker_t *s;
 
-    char         name[JK_SHM_STR_SIZ+1];
+    char         name[JK_SHM_STR_SIZ];
     /* Sequence counter starting at 0 and increasing
      * every time we change the config
      */
@@ -288,12 +299,16 @@ struct ajp_worker
 
     JK_CRIT_SEC cs;
 
-    struct sockaddr_in worker_inet_addr;    /* Contains host and port */
+    jk_sockaddr_t worker_inet_addr;    /* Contains host and port */
+    jk_sockaddr_t worker_source_inet_addr; /* Contains source ip */
     unsigned connect_retry_attempts;
-    char host[JK_SHM_STR_SIZ+1];
+    char host[JK_SHM_STR_SIZ];
     int port;
+    char source[JK_SHM_STR_SIZ];
     int addr_sequence;  /* Whether the address is resolved */
     int maintain_time;
+    int prefer_ipv6;
+
     /*
      * Open connections cache...
      *
@@ -359,11 +374,13 @@ struct ajp_worker
 
     int retry_interval;            /*  Number of milliseconds to sleep before doing a retry */
 
+    int busy_limit;                /*  Maximum allowed number of concurrent requests (if > 0) */
+
     /*
      * HTTP status that will cause failover (0 means disabled)
      */
      unsigned int http_status_fail_num;
-     int http_status_fail[JK_MAX_HTTP_STATUS_FAILS];
+     int *http_status_fail;
 };
 
 
@@ -454,8 +471,11 @@ int ajp_connection_tcp_send_message(ajp_endpoint_t * ae,
 int ajp_connection_tcp_get_message(ajp_endpoint_t * ae,
                                    jk_msg_buf_t *msg, jk_logger_t *l);
 
-int JK_METHOD ajp_maintain(jk_worker_t *pThis, time_t now, jk_logger_t *l);
+int JK_METHOD ajp_maintain(jk_worker_t *pThis, time_t now, int global, jk_logger_t *l);
 
+int JK_METHOD ajp_shutdown(jk_worker_t *pThis, jk_logger_t *l);
+
+void jk_ajp_get_cping_text(int mode, char *buf);
 int jk_ajp_get_cping_mode(const char *m, int def);
 
 int ajp_has_endpoint(jk_worker_t *pThis, jk_logger_t *l);
