@@ -23,7 +23,7 @@
  * @package jk_connect
  * @author      Gal Shachor <shachor@il.ibm.com>
  * @author      Mladen Turk <mturk@apache.org>
- * @version     $Revision: 1305771 $
+ * @version     $Revision: 1658173 $
  */
 
 
@@ -59,6 +59,29 @@ static apr_pool_t *jk_apr_pool = NULL;
 #define USE_SOCK_CLOEXEC
 #endif
 
+#ifndef INET6_ADDRSTRLEN
+/* Maximum size of an IPv6 address in ASCII */
+#define INET6_ADDRSTRLEN 46
+#endif
+
+/* 2 IPv6 adresses of length (INET6_ADDRSTRLEN-1)
+ * each suffixed with a ":" and a port (5 digits)
+ * plus " -> " plus terminating "\0"
+ */
+#define DUMP_SINFO_BUF_SZ (2 * (INET6_ADDRSTRLEN - 1 + 1 + 5) + 4 + 1)
+
+#ifndef IN6ADDRSZ
+#define IN6ADDRSZ   16
+#endif
+
+#ifndef INT16SZ
+#define INT16SZ     sizeof(short)
+#endif
+
+#if !defined(EAFNOSUPPORT) && defined(WSAEAFNOSUPPORT)
+#define EAFNOSUPPORT WSAEAFNOSUPPORT
+#endif
+
 /* our compiler cant deal with char* <-> const char* ... */
 #if defined(NETWARE) && !defined(__NOVELL_LIBC__)
 typedef char* SET_TYPE;
@@ -140,6 +163,7 @@ static int sononblock(jk_sock_t sd)
 /** Non-blocking socket connect
  * @param sd       socket to connect
  * @param addr     address to connect to
+ * @param source   optional source address
  * @param timeout  connect timeout in seconds
  *                 (<=0: no timeout=blocking)
  * @param l        logger
@@ -148,14 +172,24 @@ static int sononblock(jk_sock_t sd)
  *                               during blocking connect
  *                 0: success
  */
-static int nb_connect(jk_sock_t sd, struct sockaddr *addr, int timeout, jk_logger_t *l)
+static int nb_connect(jk_sock_t sd, jk_sockaddr_t *addr, jk_sockaddr_t *source,
+                      int timeout, jk_logger_t *l)
 {
     int rc;
+    char buf[64];
 
     JK_TRACE_ENTER(l);
 
+    if (source != NULL) {
+        if (bind(sd, (const struct sockaddr *)&source->sa.sin, source->salen)) {
+            JK_GET_SOCKET_ERRNO();
+            jk_log(l, JK_LOG_ERROR,
+                   "error during source bind on socket %d [%s] (errno=%d)",
+                   sd, jk_dump_hinfo(source, buf, sizeof(buf)), errno);
+        }
+    }
     if (timeout <= 0) {
-        rc = connect(sd, addr, sizeof(struct sockaddr_in));
+        rc = connect(sd, (const struct sockaddr *)&addr->sa.sin, addr->salen);
         JK_TRACE_EXIT(l);
         return rc;
     }
@@ -164,7 +198,7 @@ static int nb_connect(jk_sock_t sd, struct sockaddr *addr, int timeout, jk_logge
         JK_TRACE_EXIT(l);
         return -1;
     }
-    if (JK_IS_SOCKET_ERROR(connect(sd, addr, sizeof(struct sockaddr_in)))) {
+    if (JK_IS_SOCKET_ERROR(connect(sd, (const struct sockaddr *)&addr->sa.sin, addr->salen))) {
         struct timeval tv;
         fd_set wfdset, efdset;
 
@@ -214,18 +248,29 @@ static int nb_connect(jk_sock_t sd, struct sockaddr *addr, int timeout, jk_logge
 /** Non-blocking socket connect
  * @param sd       socket to connect
  * @param addr     address to connect to
+ * @param source   optional source address
  * @param timeout  connect timeout in seconds
  *                 (<=0: no timeout=blocking)
  * @param l        logger
  * @return         -1: some kind of error occured
  *                 0: success
  */
-static int nb_connect(jk_sock_t sd, struct sockaddr *addr, int timeout, jk_logger_t *l)
+static int nb_connect(jk_sock_t sd, jk_sockaddr_t *addr, jk_sockaddr_t *source,
+                      int timeout, jk_logger_t *l)
 {
     int rc = 0;
+    char buf[64];
 
     JK_TRACE_ENTER(l);
 
+    if (source != NULL) {
+        if (bind(sd, (const struct sockaddr *)&source->sa.sin, source->salen)) {
+            JK_GET_SOCKET_ERRNO();
+            jk_log(l, JK_LOG_ERROR,
+                   "error during source bind on socket %d [%s] (errno=%d)",
+                   sd, jk_dump_hinfo(source, buf, sizeof(buf)), errno);
+        }
+    }
     if (timeout > 0) {
         if (sononblock(sd)) {
             JK_TRACE_EXIT(l);
@@ -233,7 +278,7 @@ static int nb_connect(jk_sock_t sd, struct sockaddr *addr, int timeout, jk_logge
         }
     }
     do {
-        rc = connect(sd, addr, sizeof(struct sockaddr_in));
+        rc = connect(sd, (const struct sockaddr *)&addr->sa.sin, addr->salen);
     } while (rc == -1 && errno == EINTR);
 
     if ((rc == -1) && (errno == EINPROGRESS || errno == EALREADY)
@@ -278,18 +323,29 @@ static int nb_connect(jk_sock_t sd, struct sockaddr *addr, int timeout, jk_logge
 /** Non-blocking socket connect
  * @param sd       socket to connect
  * @param addr     address to connect to
+ * @param source   optional source address
  * @param timeout  connect timeout in seconds (ignored!)
  * @param l        logger
  * @return         -1: some kind of error occured
  *                 0: success
  */
-static int nb_connect(jk_sock_t sd, struct sockaddr *addr, int timeout, jk_logger_t *l)
+static int nb_connect(jk_sock_t sd, jk_sockaddr_t *addr, jk_sockaddr_t *source,
+                      int timeout, jk_logger_t *l)
 {
     int rc;
+    char buf[64];
 
     JK_TRACE_ENTER(l);
 
-    rc = connect(sd, addr, sizeof(struct sockaddr_in));
+    if (source != NULL) {
+        if (bind(sd, (const struct sockaddr *)&source->sa.sin, source->salen)) {
+            JK_GET_SOCKET_ERRNO();
+            jk_log(l, JK_LOG_ERROR,
+                   "error during source bind on socket %d [%s] (errno=%d)",
+                   sd, jk_dump_hinfo(source, buf, sizeof(buf)), errno);
+        }
+    }
+    rc = connect(sd, (const struct sockaddr *)&addr->sa.sin, addr->salen);
     JK_TRACE_EXIT(l);
     return rc;
 }
@@ -316,6 +372,20 @@ in_addr_t jk_inet_addr(const char * addrstr)
 
 #endif
 
+/** Clone a jk_sockaddr_t
+ * @param out     The source structure
+ * @param in      The target structure
+ */
+void jk_clone_sockaddr(jk_sockaddr_t *out, jk_sockaddr_t *in)
+{
+    memcpy(out, in, sizeof(*in));
+    /* The ipaddr_ptr member points to memory inside the struct.
+     * Do not copy the pointer but use the same offset relative
+     * to the struct start
+     */
+    out->ipaddr_ptr = (char *)out + ((char *)in->ipaddr_ptr - (char *)in);
+}
+
 /** Resolve the host IP
  * @param host     host or ip address
  * @param port     port
@@ -324,32 +394,24 @@ in_addr_t jk_inet_addr(const char * addrstr)
  * @return         JK_FALSE: some kind of error occured
  *                 JK_TRUE: success
  */
-int jk_resolve(const char *host, int port, struct sockaddr_in *rc,
-               void *pool, jk_logger_t *l)
+int jk_resolve(const char *host, int port, jk_sockaddr_t *saddr,
+               void *pool, int prefer_ipv6, jk_logger_t *l)
 {
-    int x;
-    struct in_addr laddr;
+    int family = JK_INET;
+    struct in_addr iaddr;
 
     JK_TRACE_ENTER(l);
 
-    memset(rc, 0, sizeof(struct sockaddr_in));
+    memset(saddr, 0, sizeof(jk_sockaddr_t));
+    if (*host >= '0' && *host <= '9' && strspn(host, "0123456789.") == strlen(host)) {
 
-    rc->sin_port = htons((short)port);
-    rc->sin_family = AF_INET;
-
-    /* Check if we only have digits in the string */
-    for (x = 0; host[x] != '\0'; x++) {
-        if (!jk_isdigit(host[x]) && host[x] != '.') {
-            break;
-        }
+        /* If we found only digits we use inet_addr() */
+        iaddr.s_addr = jk_inet_addr(host);
+        memcpy(&(saddr->sa.sin.sin_addr), &iaddr, sizeof(struct in_addr));
     }
-
-    /* If we found also characters we should make name to IP resolution */
-    if (host[x] != '\0') {
-
+    else {
 #ifdef HAVE_APR
         apr_sockaddr_t *remote_sa, *temp_sa;
-        char *remote_ipaddr;
 
         if (!jk_apr_pool) {
             if (apr_pool_create(&jk_apr_pool, (apr_pool_t *)pool) != APR_SUCCESS) {
@@ -358,55 +420,182 @@ int jk_resolve(const char *host, int port, struct sockaddr_in *rc,
             }
         }
         apr_pool_clear(jk_apr_pool);
-        if (apr_sockaddr_info_get
-            (&remote_sa, host, APR_UNSPEC, (apr_port_t) port, 0, jk_apr_pool)
-            != APR_SUCCESS) {
+        if (apr_sockaddr_info_get(&remote_sa, host, APR_UNSPEC, (apr_port_t)port,
+                                  0, jk_apr_pool) != APR_SUCCESS) {
             JK_TRACE_EXIT(l);
             return JK_FALSE;
         }
 
-        /* Since we are only handling AF_INET (IPV4) address (in_addr_t) */
-        /* make sure we find one of those.                               */
-        temp_sa = remote_sa;
-        while ((NULL != temp_sa) && (AF_INET != temp_sa->family))
-            temp_sa = temp_sa->next;
-
-        /* if temp_sa is set, we have a valid address otherwise, just return */
-        if (NULL != temp_sa) {
-            remote_sa = temp_sa;
+        /* Check if we have multiple address matches
+         */
+        if (remote_sa->next) {
+            /* Since we are only handling JK_INET (IPV4) address (in_addr_t) */
+            /* make sure we find one of those.                               */
+            temp_sa = remote_sa;
+#if APR_HAVE_IPV6
+            if (prefer_ipv6) {
+                while ((NULL != temp_sa) && (APR_INET6 != temp_sa->family))
+                    temp_sa = temp_sa->next;
+            }
+#endif
+            if (NULL != temp_sa) {
+                remote_sa = temp_sa;
+            }
+            else {
+                while ((NULL != temp_sa) && (APR_INET != temp_sa->family))
+                    temp_sa = temp_sa->next;
+#if APR_HAVE_IPV6
+                if (NULL == temp_sa) {
+                    temp_sa = remote_sa;
+                    while ((NULL != temp_sa) && (APR_INET6 != temp_sa->family))
+                        temp_sa = temp_sa->next;
+                }
+#endif
+            }
+            /* if temp_sa is set, we have a valid address otherwise, just return */
+            if (NULL != temp_sa) {
+                remote_sa = temp_sa;
+            }
+            else {
+                JK_TRACE_EXIT(l);
+                return JK_FALSE;
+            }
         }
+        if (remote_sa->family == APR_INET) {
+            saddr->sa.sin = remote_sa->sa.sin;
+            family = JK_INET;
+        }
+#if APR_HAVE_IPV6
         else {
+            saddr->sa.sin6 = remote_sa->sa.sin6;
+            family = JK_INET6;
+        }
+#endif
+#else /* HAVE_APR */
+        /* Without APR go the classic way.
+         */
+#if defined(HAVE_GETADDRINFO)
+        /* TODO:
+         * 1. Check for numeric IPV6 addresses
+         * 2. Do we need to set service name for getaddrinfo?
+         */
+        struct addrinfo hints, *ai_list, *ai = NULL;
+        int error;
+        char  pbuf[12];
+        char *pbufptr = NULL;
+
+        memset(&hints, 0, sizeof(hints));
+        hints.ai_socktype = SOCK_STREAM;
+        hints.ai_protocol = IPPROTO_TCP;
+        
+#if JK_HAVE_IPV6
+        if (strchr(host, ':')) {
+            /* If host name contains collon this must be IPV6 address.
+             * Set prefer_ipv6 flag in this case if it wasn't set already
+             */
+            prefer_ipv6 = 1;            
+        }
+        if (prefer_ipv6)
+            hints.ai_family = JK_INET6;
+        else
+#endif
+            hints.ai_family = JK_INET;
+        if (port > 0) {
+            snprintf(pbuf, sizeof(pbuf), "%d", port);
+            pbufptr = pbuf;
+        }
+        error = getaddrinfo(host, pbufptr, &hints, &ai_list);
+#if JK_HAVE_IPV6
+        /* XXX:
+         * Is the check for EAI_FAMILY/WSAEAFNOSUPPORT correct
+         * way to retry the IPv4 address?
+         */
+        if (error == EAI_FAMILY && prefer_ipv6) {
+            hints.ai_family = JK_INET;
+            error = getaddrinfo(host, pbufptr, &hints, &ai_list);
+        }
+#endif
+        if (error) {
             JK_TRACE_EXIT(l);
+            errno = error;
             return JK_FALSE;
         }
-
-        apr_sockaddr_ip_get(&remote_ipaddr, remote_sa);
-
-        laddr.s_addr = jk_inet_addr(remote_ipaddr);
-
-#else /* HAVE_APR */
+#if JK_HAVE_IPV6
+        if (prefer_ipv6) {
+            ai = ai_list;
+            while (ai) {
+                if (ai->ai_family == JK_INET6) {
+                    /* ignore elements without required address info */
+                    if((ai->ai_addr != NULL) && (ai->ai_addrlen > 0)) {                        
+                        family = JK_INET6;
+                        break;
+                    }
+                }
+                ai = ai->ai_next;
+            }
+        }
+#endif
+        if (ai == NULL) {
+            ai = ai_list;
+            while (ai) {
+                if (ai->ai_family == JK_INET) {
+                    /* ignore elements without required address info */
+                    if((ai->ai_addr != NULL) && (ai->ai_addrlen > 0)) {                        
+                        family = JK_INET;
+                        break;
+                    }
+                }
+                ai = ai->ai_next;
+            }
+        }
+        if (ai == NULL) {
+            /* No address found
+             * XXX: Use better error code?
+             */
+            freeaddrinfo(ai_list);
+            JK_TRACE_EXIT(l);
+            errno = ENOENT;
+            return JK_FALSE;
+        }
+        memcpy(&(saddr->sa), ai->ai_addr, ai->ai_addrlen);
+        freeaddrinfo(ai_list);
+#else /* HAVE_GETADDRINFO */
+        struct hostent *hoste;
 
         /* XXX : WARNING : We should really use gethostbyname_r in multi-threaded env */
         /* Fortunatly when APR is available, ie under Apache 2.0, we use it */
 #if defined(NETWARE) && !defined(__NOVELL_LIBC__)
-        struct hostent *hoste = gethostbyname((char*)host);
+        hoste = gethostbyname((char*)host);
 #else
-        struct hostent *hoste = gethostbyname(host);
+        hoste = gethostbyname(host);
 #endif
         if (!hoste) {
             JK_TRACE_EXIT(l);
             return JK_FALSE;
         }
-
-        laddr = *((struct in_addr *)hoste->h_addr_list[0]);
-
+        iaddr = *((struct in_addr *)hoste->h_addr_list[0]);
+        memcpy(&(saddr->sa.sin.sin_addr), &iaddr, sizeof(struct in_addr));
+#endif /* HAVE_GETADDRINFO */
 #endif /* HAVE_APR */
     }
-    else {
-        /* If we found only digits we use inet_addr() */
-        laddr.s_addr = jk_inet_addr(host);
+
+    if (family == JK_INET) {
+        saddr->ipaddr_ptr = &(saddr->sa.sin.sin_addr);
+        saddr->ipaddr_len = (int)sizeof(struct in_addr);
+        saddr->salen      = (int)sizeof(struct sockaddr_in);
     }
-    memcpy(&(rc->sin_addr), &laddr, sizeof(laddr));
+#if JK_HAVE_IPV6
+    else {
+        saddr->ipaddr_ptr = &(saddr->sa.sin6.sin6_addr);
+        saddr->ipaddr_len = (int)sizeof(struct in6_addr);
+        saddr->salen      = (int)sizeof(struct sockaddr_in6);
+    }
+#endif
+    saddr->sa.sin.sin_family = family;
+    /* XXX IPv6: assumes sin_port and sin6_port at same offset */
+    saddr->sa.sin.sin_port = htons(port);
+    saddr->port   = port;
+    saddr->family = family;
 
     JK_TRACE_EXIT(l);
     return JK_TRUE;
@@ -424,11 +613,12 @@ int jk_resolve(const char *host, int port, struct sockaddr_in *rc,
  *                  created socket: success
  * @remark          Cares about errno
  */
-jk_sock_t jk_open_socket(struct sockaddr_in *addr, int keepalive,
+jk_sock_t jk_open_socket(jk_sockaddr_t *addr, jk_sockaddr_t *source,
+                         int keepalive,
                          int timeout, int connect_timeout,
                          int sock_buf, jk_logger_t *l)
 {
-    char buf[64];
+    char buf[DUMP_SINFO_BUF_SZ];
     jk_sock_t sd;
     int set = 1;
     int ret = 0;
@@ -443,7 +633,7 @@ jk_sock_t jk_open_socket(struct sockaddr_in *addr, int keepalive,
 #if defined(SOCK_CLOEXEC) && defined(USE_SOCK_CLOEXEC)
     flags |= SOCK_CLOEXEC;
 #endif
-    sd = socket(AF_INET, SOCK_STREAM | flags, 0);
+    sd = socket(addr->family, SOCK_STREAM | flags, 0);
     if (!IS_VALID_SOCKET(sd)) {
         JK_GET_SOCKET_ERRNO();
         jk_log(l, JK_LOG_ERROR,
@@ -608,14 +798,14 @@ jk_sock_t jk_open_socket(struct sockaddr_in *addr, int keepalive,
     if (JK_IS_DEBUG_LEVEL(l))
         jk_log(l, JK_LOG_DEBUG,
                 "trying to connect socket %d to %s", sd,
-                jk_dump_hinfo(addr, buf));
+                jk_dump_hinfo(addr, buf, sizeof(buf)));
 
 /* Need more infos for BSD 4.4 and Unix 98 defines, for now only
-iSeries when Unix98 is required at compil time */
+iSeries when Unix98 is required at compile time */
 #if (_XOPEN_SOURCE >= 520) && defined(AS400)
-    ((struct sockaddr *)addr)->sa_len = sizeof(struct sockaddr_in);
+    ((struct sockaddr *)addr)->sa.sin.sa_len = sizeof(struct sockaddr_in);
 #endif
-    ret = nb_connect(sd, (struct sockaddr *)addr, connect_timeout, l);
+    ret = nb_connect(sd, addr, source, connect_timeout, l);
 #if defined(WIN32) || (defined(NETWARE) && defined(__NOVELL_LIBC__))
     if (JK_IS_SOCKET_ERROR(ret)) {
         JK_GET_SOCKET_ERRNO();
@@ -626,14 +816,14 @@ iSeries when Unix98 is required at compil time */
     if (ret) {
         jk_log(l, JK_LOG_INFO,
                "connect to %s failed (errno=%d)",
-               jk_dump_hinfo(addr, buf), errno);
+               jk_dump_hinfo(addr, buf, sizeof(buf)), errno);
         jk_close_socket(sd, l);
         sd = JK_INVALID_SOCKET;
     }
     else {
         if (JK_IS_DEBUG_LEVEL(l))
             jk_log(l, JK_LOG_DEBUG, "socket %d [%s] connected",
-                   sd, jk_dump_sinfo(sd, buf));
+                   sd, jk_dump_sinfo(sd, buf, sizeof(buf)));
     }
     JK_TRACE_EXIT(l);
     return sd;
@@ -675,8 +865,18 @@ int jk_close_socket(jk_sock_t sd, jk_logger_t *l)
 #ifndef MAX_SECS_TO_LINGER
 #define MAX_SECS_TO_LINGER 2
 #endif
-#define MS_TO_LINGER  500
-#define MS_TO_LINGER_LAST 2
+
+#ifndef MS_TO_LINGER
+#define MS_TO_LINGER  100
+#endif
+
+#ifndef MS_TO_LINGER_LAST
+#define MS_TO_LINGER_LAST 20
+#endif
+
+#ifndef MAX_READ_RETRY
+#define MAX_READ_RETRY 10
+#endif
 
 #ifndef MAX_LINGER_BYTES
 #define MAX_LINGER_BYTES 32768
@@ -710,7 +910,7 @@ int jk_close_socket(jk_sock_t sd, jk_logger_t *l)
 int jk_shutdown_socket(jk_sock_t sd, jk_logger_t *l)
 {
     char dummy[512];
-    char buf[64];
+    char buf[DUMP_SINFO_BUF_SZ];
     char *sb = NULL;
     int rc = 0;
     size_t rd = 0;
@@ -728,7 +928,7 @@ int jk_shutdown_socket(jk_sock_t sd, jk_logger_t *l)
 
     save_errno = errno;
     if (JK_IS_DEBUG_LEVEL(l)) {
-        sb = jk_dump_sinfo(sd, buf);
+        sb = jk_dump_sinfo(sd, buf, sizeof(buf));
         jk_log(l, JK_LOG_DEBUG, "About to shutdown socket %d [%s]",
                sd, sb);
     }
@@ -752,7 +952,9 @@ int jk_shutdown_socket(jk_sock_t sd, jk_logger_t *l)
             /* Do a restartable read on the socket
              * draining out all the data currently in the socket buffer.
              */
+            int num = 0;
             do {
+                num++;
 #if defined(WIN32) || (defined(NETWARE) && defined(__NOVELL_LIBC__))
                 rc = recv(sd, &dummy[0], sizeof(dummy), 0);
                 if (JK_IS_SOCKET_ERROR(rc))
@@ -762,7 +964,7 @@ int jk_shutdown_socket(jk_sock_t sd, jk_logger_t *l)
 #endif
                 if (rc > 0)
                     rp += rc;
-            } while (JK_IS_SOCKET_ERROR(rc) && (errno == EINTR || errno == EAGAIN));
+            } while (JK_IS_SOCKET_ERROR(rc) && (errno == EINTR || errno == EAGAIN) && num < MAX_READ_RETRY);
 
             if (rc < 0) {
                 /* Read failed.
@@ -780,7 +982,7 @@ int jk_shutdown_socket(jk_sock_t sd, jk_logger_t *l)
         rd += rp;
         if (rp < sizeof(dummy)) {
             if (timeout > MS_TO_LINGER_LAST) {
-                /* Try one last time with a short timeout 
+                /* Try one last time with a short timeout
                 */
                 timeout = MS_TO_LINGER_LAST;
                 continue;
@@ -922,45 +1124,227 @@ int jk_tcp_socket_recvfull(jk_sock_t sd, unsigned char *b, int len, jk_logger_t
     return rdlen;
 }
 
+/* const char *
+ * inet_ntop4(src, dst, size)
+ *  format an IPv4 address, more or less like inet_ntoa()
+ * return:
+ *  `dst' (as a const)
+ * notes:
+ *  (1) uses no statics
+ *  (2) takes a u_char* not an in_addr as input
+ * author:
+ *  Paul Vixie, 1996.
+ */
+static const char *inet_ntop4(const unsigned char *src, char *dst, size_t size)
+{
+    const size_t MIN_SIZE = 16; /* space for 255.255.255.255\0 */
+    int n = 0;
+    char *next = dst;
+
+    if (size < MIN_SIZE) {
+        errno = ENOSPC;
+        return NULL;
+    }
+    do {
+        unsigned char u = *src++;
+        if (u > 99) {
+            *next++ = '0' + u/100;
+             u %= 100;
+            *next++ = '0' + u/10;
+             u %= 10;
+        }
+        else if (u > 9) {
+            *next++ = '0' + u/10;
+             u %= 10;
+        }
+        *next++ = '0' + u;
+        *next++ = '.';
+        n++;
+    } while (n < 4);
+    *--next = '\0';
+    return dst;
+}
+
+#if JK_HAVE_IPV6
+/* const char *
+ * inet_ntop6(src, dst, size)
+ *  convert IPv6 binary address into presentation (printable) format
+ * author:
+ *  Paul Vixie, 1996.
+ */
+static const char *inet_ntop6(const unsigned char *src, char *dst, size_t size)
+{
+    /*
+     * Note that int32_t and int16_t need only be "at least" large enough
+     * to contain a value of the specified size.  On some systems, like
+     * Crays, there is no such thing as an integer variable with 16 bits.
+     * Keep this in mind if you think this function should have been coded
+     * to use pointer overlays.  All the world's not a VAX.
+     */
+    char tmp[INET6_ADDRSTRLEN], *tp;
+    struct { int base, len; } best = {-1, 0}, cur = {-1, 0};
+    unsigned int words[IN6ADDRSZ / INT16SZ];
+    int i;
+    const unsigned char *next_src, *src_end;
+    unsigned int *next_dest;
+
+    /*
+     * Preprocess:
+     *  Copy the input (bytewise) array into a wordwise array.
+     *  Find the longest run of 0x00's in src[] for :: shorthanding.
+     */
+    next_src = src;
+    src_end = src + IN6ADDRSZ;
+    next_dest = words;
+    i = 0;
+    do {
+        unsigned int next_word = (unsigned int)*next_src++;
+        next_word <<= 8;
+        next_word |= (unsigned int)*next_src++;
+        *next_dest++ = next_word;
+
+        if (next_word == 0) {
+            if (cur.base == -1) {
+                cur.base = i;
+                cur.len = 1;
+            }
+            else {
+                cur.len++;
+            }
+        } else {
+            if (cur.base != -1) {
+                if (best.base == -1 || cur.len > best.len) {
+                    best = cur;
+                }
+                cur.base = -1;
+            }
+        }
+
+        i++;
+    } while (next_src < src_end);
+
+    if (cur.base != -1) {
+        if (best.base == -1 || cur.len > best.len) {
+            best = cur;
+        }
+    }
+    if (best.base != -1 && best.len < 2) {
+        best.base = -1;
+    }
+
+    /*
+     * Format the result.
+     */
+    tp = tmp;
+    for (i = 0; i < (IN6ADDRSZ / INT16SZ);) {
+        /* Are we inside the best run of 0x00's? */
+        if (i == best.base) {
+            *tp++ = ':';
+            i += best.len;
+            continue;
+        }
+        /* Are we following an initial run of 0x00s or any real hex? */
+        if (i != 0) {
+            *tp++ = ':';
+        }
+        /* Is this address an encapsulated IPv4? */
+        if (i == 6 && best.base == 0 &&
+            (best.len == 6 || (best.len == 5 && words[5] == 0xffff))) {
+            if (!inet_ntop4(src+12, tp, sizeof tmp - (tp - tmp))) {
+                return (NULL);
+            }
+            tp += strlen(tp);
+            break;
+        }
+        tp += sprintf(tp, "%x", words[i]);
+        i++;
+    }
+    /* Was it a trailing run of 0x00's? */
+    if (best.base != -1 && (best.base + best.len) == (IN6ADDRSZ / INT16SZ)) {
+        *tp++ = ':';
+    }
+    *tp++ = '\0';
+
+    /*
+     * Check for overflow, copy, and we're done.
+     */
+    if ((size_t)(tp - tmp) > size) {
+        errno = ENOSPC;
+        return NULL;
+    }
+    strcpy(dst, tmp);
+    return dst;
+}
+#endif
+
 /**
- * dump a sockaddr_in in A.B.C.D:P in ASCII buffer
+ * dump a jk_sockaddr_t in A.B.C.D:P in ASCII buffer
  *
  */
-char *jk_dump_hinfo(struct sockaddr_in *saddr, char *buf)
+char *jk_dump_hinfo(jk_sockaddr_t *saddr, char *buf, size_t size)
 {
-    unsigned long laddr = (unsigned long)htonl(saddr->sin_addr.s_addr);
-    unsigned short lport = (unsigned short)htons(saddr->sin_port);
+    char pb[8];
 
-    sprintf(buf, "%d.%d.%d.%d:%d",
-            (int)(laddr >> 24), (int)((laddr >> 16) & 0xff),
-            (int)((laddr >> 8) & 0xff), (int)(laddr & 0xff), (int)lport);
+    if (saddr->family == JK_INET) {
+        inet_ntop4(saddr->ipaddr_ptr, buf, size);
+    }
+#if JK_HAVE_IPV6
+    else {
+        inet_ntop6(saddr->ipaddr_ptr, buf, size);
+    }
+#endif
+    sprintf(pb, ":%d", saddr->port);
 
+    strncat(buf, pb, size - strlen(buf) - 1);
     return buf;
 }
 
-char *jk_dump_sinfo(jk_sock_t sd, char *buf)
+char *jk_dump_sinfo(jk_sock_t sd, char *buf, size_t size)
 {
-    struct sockaddr_in rsaddr;
-    struct sockaddr_in lsaddr;
-    socklen_t          salen;
+    struct sockaddr rsaddr;
+    struct sockaddr lsaddr;
+    socklen_t       salen;
 
     salen = sizeof(struct sockaddr);
-    if (getsockname(sd, (struct sockaddr *)&lsaddr, &salen) == 0) {
+    if (getsockname(sd, &lsaddr, &salen) == 0) {
         salen = sizeof(struct sockaddr);
-        if (getpeername(sd, (struct sockaddr *)&rsaddr, &salen) == 0) {
-            unsigned long  laddr = (unsigned  long)htonl(lsaddr.sin_addr.s_addr);
-            unsigned short lport = (unsigned short)htons(lsaddr.sin_port);
-            unsigned long  raddr = (unsigned  long)htonl(rsaddr.sin_addr.s_addr);
-            unsigned short rport = (unsigned short)htons(rsaddr.sin_port);
-            sprintf(buf, "%d.%d.%d.%d:%d -> %d.%d.%d.%d:%d",
-                    (int)(laddr >> 24), (int)((laddr >> 16) & 0xff),
-                    (int)((laddr >> 8) & 0xff), (int)(laddr & 0xff), (int)lport,
-                    (int)(raddr >> 24), (int)((raddr >> 16) & 0xff),
-                    (int)((raddr >> 8) & 0xff), (int)(raddr & 0xff), (int)rport);
+        if (getpeername(sd, &rsaddr, &salen) == 0) {
+            char   pb[8];
+            size_t ps;
+            if (lsaddr.sa_family == JK_INET) {
+                struct sockaddr_in  *sa = (struct sockaddr_in  *)&lsaddr;
+                inet_ntop4((unsigned char *)&sa->sin_addr, buf, size);
+                sprintf(pb, ":%d", (unsigned int)htons(sa->sin_port));
+            }
+#if JK_HAVE_IPV6
+            else {
+                struct sockaddr_in6 *sa = (struct sockaddr_in6 *)&lsaddr;
+                inet_ntop6((unsigned char *)&sa->sin6_addr, buf, size);
+                sprintf(pb, ":%d", (unsigned int)htons(sa->sin6_port));
+            }
+#endif
+            ps = strlen(buf);
+            strncat(buf, pb, size - ps - 1);
+            ps = strlen(buf);
+            strncat(buf, " -> ", size - ps - 1);
+            ps = strlen(buf);
+            if (rsaddr.sa_family == JK_INET) {
+                struct sockaddr_in  *sa = (struct sockaddr_in  *)&rsaddr;
+                inet_ntop4((unsigned char *)&sa->sin_addr,  buf + ps, size - ps);
+                sprintf(pb, ":%d", (unsigned int)htons(sa->sin_port));
+            }
+#if JK_HAVE_IPV6
+            else {
+                struct sockaddr_in6 *sa = (struct sockaddr_in6 *)&rsaddr;
+                inet_ntop6((unsigned char *)&sa->sin6_addr, buf + ps, size - ps);
+                sprintf(pb, ":%d", (unsigned int)htons(sa->sin6_port));
+            }
+#endif
+            strncat(buf, pb, size - strlen(buf) - 1);
             return buf;
         }
     }
-    sprintf(buf, "errno=%d", errno);
+    snprintf(buf, size, "errno=%d", errno);
     return buf;
 }
 
@@ -981,7 +1365,7 @@ int jk_is_input_event(jk_sock_t sd, int timeout, jk_logger_t *l)
     struct pollfd fds;
     int rc;
     int save_errno;
-    char buf[64];
+    char buf[DUMP_SINFO_BUF_SZ];
 
     JK_TRACE_ENTER(l);
 
@@ -998,7 +1382,7 @@ int jk_is_input_event(jk_sock_t sd, int timeout, jk_logger_t *l)
         if (JK_IS_DEBUG_LEVEL(l)) {
             jk_log(l, JK_LOG_DEBUG,
                    "timeout during poll on socket %d [%s] (timeout=%d)",
-                   sd, jk_dump_sinfo(sd, buf), timeout);
+                   sd, jk_dump_sinfo(sd, buf, sizeof(buf)), timeout);
         }
         /* Timeout. Set the errno to timeout */
         errno = ETIMEDOUT;
@@ -1010,7 +1394,7 @@ int jk_is_input_event(jk_sock_t sd, int timeout, jk_logger_t *l)
         if (JK_IS_DEBUG_LEVEL(l)) {
             jk_log(l, JK_LOG_DEBUG,
                    "error during poll on socket %d [%s] (errno=%d)",
-                   sd, jk_dump_sinfo(sd, buf), errno);
+                   sd, jk_dump_sinfo(sd, buf, sizeof(buf)), errno);
         }
         errno = save_errno;
         JK_TRACE_EXIT(l);
@@ -1021,7 +1405,7 @@ int jk_is_input_event(jk_sock_t sd, int timeout, jk_logger_t *l)
         if (JK_IS_DEBUG_LEVEL(l)) {
             jk_log(l, JK_LOG_DEBUG,
                    "error event during poll on socket %d [%s] (event=%d)",
-                   sd, jk_dump_sinfo(sd, buf), save_errno);
+                   sd, jk_dump_sinfo(sd, buf, sizeof(buf)), save_errno);
         }
         errno = save_errno;
         JK_TRACE_EXIT(l);
@@ -1038,7 +1422,7 @@ int jk_is_input_event(jk_sock_t sd, int timeout, jk_logger_t *l)
     struct timeval tv;
     int rc;
     int save_errno;
-    char buf[64];
+    char buf[DUMP_SINFO_BUF_SZ];
 
     JK_TRACE_ENTER(l);
 
@@ -1056,7 +1440,7 @@ int jk_is_input_event(jk_sock_t sd, int timeout, jk_logger_t *l)
         if (JK_IS_DEBUG_LEVEL(l)) {
             jk_log(l, JK_LOG_DEBUG,
                    "timeout during select on socket %d [%s] (timeout=%d)",
-                   sd, jk_dump_sinfo(sd, buf), timeout);
+                   sd, jk_dump_sinfo(sd, buf, sizeof(buf)), timeout);
         }
         /* Timeout. Set the errno to timeout */
 #if defined(WIN32) || (defined(NETWARE) && defined(__NOVELL_LIBC__))
@@ -1072,7 +1456,7 @@ int jk_is_input_event(jk_sock_t sd, int timeout, jk_logger_t *l)
         if (JK_IS_DEBUG_LEVEL(l)) {
             jk_log(l, JK_LOG_DEBUG,
                    "error during select on socket %d [%s] (errno=%d)",
-                   sd, jk_dump_sinfo(sd, buf), errno);
+                   sd, jk_dump_sinfo(sd, buf, sizeof(buf)), errno);
         }
         errno = save_errno;
         JK_TRACE_EXIT(l);
