@@ -19,13 +19,25 @@
  * Description: Global definitions and include files that should exist     *
  *              anywhere                                                   *
  * Author:      Gal Shachor <shachor@il.ibm.com>                           *
- * Version:     $Revision: 1126561 $                                               *
+ * Version:     $Revision: 1650043 $                                               *
  ***************************************************************************/
 
 #ifndef JK_GLOBAL_H
 #define JK_GLOBAL_H
 
 #if defined(HAVE_CONFIG_H)
+
+/* Undefine autoconf defines also existing
+ * in Apache web server includes, and which
+ * will be redefined in our config.h */
+#undef PACKAGE
+#undef PACKAGE_NAME
+#undef PACKAGE_STRING
+#undef PACKAGE_TARNAME
+#undef PACKAGE_VERSION
+#undef PACKAGE_BUGREPORT
+#undef PACKAGE_URL
+
 #include "config.h"
 #endif
 
@@ -65,6 +77,7 @@
 #ifdef HAVE_APR
 #include "apr_lib.h"
 #include "apr_strings.h"
+#include "apr_atomic.h"
 #endif
 
 #include <stdio.h>
@@ -188,6 +201,8 @@ extern "C"
 #define JK_LF (10)
 #define JK_CR (13)
 
+#define JK_MAX_NAME_LEN (60)
+
 #define JK_SESSION_IDENTIFIER "JSESSIONID"
 #define JK_PATH_SESSION_IDENTIFIER ";jsessionid"
 
@@ -251,6 +266,7 @@ extern "C"
  */
 
 #define JK_OPT_FWDURIMASK           0x0007
+#define JK_OPT_FWDADDRMASK          0x0810
 
 #define JK_OPT_FWDURICOMPAT         0x0001
 #define JK_OPT_FWDURICOMPATUNPARSED 0x0002
@@ -268,8 +284,15 @@ extern "C"
 #define JK_OPT_FWDCERTCHAIN         0x0100
 #define JK_OPT_FWDKEYSIZE           0x0200
 #define JK_OPT_REJECTUNSAFE         0x0400
+/* Forward physical tcp peer address instead of
+ * client address as provided e.g. by httpd mod_remoteip. */
+#define JK_OPT_FWDPHYSICAL          0x0800
+#define JK_OPT_COLLAPSEMASK         0x7000
+#define JK_OPT_COLLAPSEALL          0x1000
+#define JK_OPT_COLLAPSENONE         0x2000
+#define JK_OPT_COLLAPSEUNMOUNT      0x4000
 
-#define JK_OPT_DEFAULT              (JK_OPT_FWDURIDEFAULT | JK_OPT_FWDKEYSIZE)
+#define JK_OPT_DEFAULT              (JK_OPT_FWDURIDEFAULT | JK_OPT_FWDKEYSIZE | JK_OPT_COLLAPSEUNMOUNT)
 
 /* Check for EBCDIC systems */
 
@@ -345,6 +368,9 @@ extern "C"
                                (tv)->tv_usec = tb.millitm * 1000; }
 #define HAVE_VSNPRINTF
 #define HAVE_SNPRINTF
+#define HAVE_SOCKADDR_STORAGE
+#define HAVE_AF_INET6
+#define HAVE_GETADDRINFO
 #ifdef HAVE_APR
 #define snprintf apr_snprintf
 #define vsnprintf apr_vsnprintf
@@ -384,7 +410,6 @@ typedef int jk_sock_t;
 #define IS_VALID_SOCKET(s) ((s) > 0)
 #define JK_INVALID_SOCKET  (-1)
 #endif
-
 #ifdef NETWARE
 #ifdef __NOVELL_LIBC__
 #define MAX_PATH 511
@@ -393,10 +418,94 @@ typedef int jk_sock_t;
 #endif
 #endif
 
+#if defined(_MSC_VER) && (_MSC_VER > 1200)
+#define stat    _stat
+#endif
+
 #ifdef AS400_UTF8
 #define strcasecmp(a,b) apr_strnatcasecmp(a,b)
 #endif
 
+/* Atomics */
+#if defined(WIN32)
+#define JK_ATOMIC_INCREMENT(x) InterlockedIncrement(x)
+#define JK_ATOMIC_DECREMENT(x) \
+    do {\
+        if (InterlockedDecrement(x) < 0) InterlockedIncrement(x);\
+    } while (0)
+#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 1))
+#define JK_ATOMIC_INCREMENT(x) __sync_add_and_fetch(x, 1)
+#define JK_ATOMIC_DECREMENT(x) \
+    do {\
+        if (__sync_sub_and_fetch(x, 1) < 0) __sync_add_and_fetch(x, 1);\
+    } while (0)
+#elif defined(HAVE_APR)
+#define JK_ATOMIC_INCREMENT(x) ((int)apr_atomic_inc32((volatile apr_uint32_t *)x) + 1)
+#define JK_ATOMIC_DECREMENT(x) \
+    do {\
+        /* apr_atomic_add32 returns the *old* value */\
+        apr_uint32_t y;\
+        y = apr_atomic_add32((volatile apr_uint32_t *)x, -1);\
+        if (y == 0 || y > INT_MAX) apr_atomic_inc32((volatile apr_uint32_t *)x);\
+    } while (0)
+#else
+#define JK_ATOMIC_MISSING
+#define JK_ATOMIC_INCREMENT(x) (++(*x))
+#define JK_ATOMIC_DECREMENT(x) \
+    do {\
+        if (--(*x) < 0) (++(*x));\
+    } while (0)
+#endif
+
+/* IPV6 support */
+#if defined(HAVE_APR)
+#define JK_HAVE_IPV6            APR_HAVE_IPV6
+#define JK_INET                 APR_INET
+#define JK_INET6                APR_INET6
+#define JK_UNSPEC               APR_UNSPEC
+#else
+#if defined(WIN32) || defined(HAVE_AF_INET6)
+#define JK_HAVE_IPV6            1
+#define JK_INET6                AF_INET6
+#else
+#define JK_HAVE_IPV6            0
+#define JK_INET6                0
+#endif
+#define JK_INET                 AF_INET
+#if defined(AF_UNSPEC)
+#define JK_UNSPEC               AF_UNSPEC
+#else
+#define JK_UNSPEC               0
+#endif
+#endif
+
+typedef struct jk_sockaddr_t jk_sockaddr_t;
+struct jk_sockaddr_t {
+    int         family;
+    int         port;
+    int         salen;
+    int         ipaddr_len;
+    /** This points to the IP address structure within the appropriate
+     *  sockaddr structure.  */
+    void       *ipaddr_ptr;
+    /** Union of either IPv4 or IPv6 sockaddr. */
+    union {
+        /** IPv4 sockaddr structure */
+        struct sockaddr_in sin;
+#if JK_HAVE_IPV6
+        /** IPv6 sockaddr structure */
+        struct sockaddr_in6 sin6;
+#endif
+        /** Placeholder to ensure that the size of this union is not
+         * dependent on whether JK_HAVE_IPV6 is defined. */
+#ifdef HAVE_SOCKADDR_STORAGE
+        struct sockaddr_storage sas;
+#else
+        char sas[128];
+#endif
+    } sa;
+};
+
 #ifdef __cplusplus
 }
 #endif                          /* __cplusplus */
