@@ -22,7 +22,7 @@
  * Author:      Mladen Turk <mturk@apache.org>                             *
  * Author:      Rainer Jung <rjung@apache.org>                             *
  * Based on:                                                               *
- * Version:     $Revision: 1343092 $                                          *
+ * Version:     $Revision: 1665454 $                                          *
  ***************************************************************************/
 
 #include "jk_pool.h"
@@ -30,7 +30,6 @@
 #include "jk_util.h"
 #include "jk_worker.h"
 #include "jk_lb_worker.h"
-#include "jk_ajp13.h"
 #include "jk_ajp13_worker.h"
 #include "jk_ajp14_worker.h"
 #include "jk_mt.h"
@@ -51,7 +50,7 @@
  *       activation is none of JK_LB_ACTIVATION_STOPPED, JK_LB_ACTIVATION_DISABLED
  */
 #define JK_WORKER_USABLE(s, activation)   ((s) <= JK_LB_STATE_FORCE && activation == JK_LB_ACTIVATION_ACTIVE)
-#define JK_WORKER_USABLE_STICKY(s, activation)   ((s) <= JK_LB_STATE_BUSY && activation != JK_LB_ACTIVATION_STOPPED)
+#define JK_WORKER_USABLE_STICKY(s, activation)   ((s) <= JK_LB_STATE_FORCE && activation != JK_LB_ACTIVATION_STOPPED)
 
 static const char *lb_locking_type[] = {
     JK_LB_LOCK_TEXT_OPTIMISTIC,
@@ -151,6 +150,14 @@ static jk_uint64_t scm(jk_uint64_t a, jk_uint64_t b)
 }
 
 /* Return the string representation of the lb lock type */
+/* based on the integer representation */
+const char *jk_lb_get_lock_direct(int lblock, jk_logger_t *l)
+{
+    return lb_locking_type[lblock];
+}
+
+/* Return the string representation of the lb lock type */
+/* based on the lb worker struct */
 const char *jk_lb_get_lock(lb_worker_t *p, jk_logger_t *l)
 {
     return lb_locking_type[p->lblock];
@@ -169,6 +176,14 @@ int jk_lb_get_lock_code(const char *v)
 }
 
 /* Return the string representation of the lb method type */
+/* based on the integer representation */
+const char *jk_lb_get_method_direct(int lbmethod, jk_logger_t *l)
+{
+    return lb_method_type[lbmethod];
+}
+
+/* Return the string representation of the lb method type */
+/* based on the lb worker struct */
 const char *jk_lb_get_method(lb_worker_t *p, jk_logger_t *l)
 {
     return lb_method_type[p->lbmethod];
@@ -193,6 +208,14 @@ int jk_lb_get_method_code(const char *v)
 }
 
 /* Return the string representation of the balance worker state */
+/* based on the integer representation */
+const char *jk_lb_get_state_direct(int state, jk_logger_t *l)
+{
+    return lb_state_type[state];
+}
+
+/* Return the string representation of the balance worker state */
+/* based on the sub worker struct */
 const char *jk_lb_get_state(lb_sub_worker_t *p, jk_logger_t *l)
 {
     return lb_state_type[p->s->state];
@@ -334,8 +357,6 @@ void jk_lb_pull(lb_worker_t *p, int locked, jk_logger_t *l)
     p->lbmethod = p->s->lbmethod;
     p->lblock = p->s->lblock;
     p->max_packet_size = p->s->max_packet_size;
-    strncpy(p->session_cookie, p->s->session_cookie, JK_SHM_STR_SIZ);
-    strncpy(p->session_path, p->s->session_path, JK_SHM_STR_SIZ);
 
     for (i = 0; i < p->num_of_workers; i++) {
         jk_lb_pull_worker(p, i, l);
@@ -370,8 +391,6 @@ void jk_lb_push(lb_worker_t *p, int locked, jk_logger_t *l)
     p->s->lbmethod = p->lbmethod;
     p->s->lblock = p->lblock;
     p->s->max_packet_size = p->max_packet_size;
-    strncpy(p->s->session_cookie, p->session_cookie, JK_SHM_STR_SIZ);
-    strncpy(p->s->session_path, p->session_path, JK_SHM_STR_SIZ);
 
     for (i = 0; i < p->num_of_workers; i++) {
         lb_sub_worker_t *w = &p->lb_workers[i];
@@ -475,6 +494,11 @@ static char *get_cookie(jk_ws_service_t *s, const char *name)
                             sz = strlen(id_start);
                             id_end = id_start + sz;
                         }
+                        /* Chop off surrounding '"' (quoted cookie) */
+                        if (sz > 1 && *id_start == '"' && *(id_start + sz - 1) == '"') {
+                            id_start++;
+                            sz -= 2;
+                        }
                         if (result == NULL) {
                             result = jk_pool_alloc(s->pool, sz + 1);
                             memcpy(result, id_start, sz);
@@ -518,10 +542,10 @@ static char *get_sessionid(jk_ws_service_t *s, lb_worker_t *p, jk_logger_t *l)
         return val;
     }
 
-    // set session_path
+    /* set session_path */
     session_path = (s->extension.session_path) ?
             s->extension.session_path : p->session_path;
-    // set session_cookie
+    /* set session_cookie */
     session_cookie = (s->extension.session_cookie) ?
             s->extension.session_cookie : p->session_cookie;
 
@@ -577,7 +601,7 @@ static int recover_workers(lb_worker_t *p,
         w = &p->lb_workers[i];
         aw = (ajp_worker_t *)w->worker->worker_private;
         if (w->s->state == JK_LB_STATE_ERROR) {
-            elapsed = (int)difftime(now, w->s->error_time);
+            elapsed = (int)difftime(now, w->s->last_error_time);
             if (elapsed <= p->recover_wait_time) {
                 if (JK_IS_DEBUG_LEVEL(l))
                     jk_log(l, JK_LOG_DEBUG,
@@ -596,8 +620,8 @@ static int recover_workers(lb_worker_t *p,
                 non_error++;
             }
         }
-        else if (w->s->error_time > 0 &&
-                 (int)difftime(now, w->s->error_time) >= p->error_escalation_time &&
+        else if (w->s->first_error_time > 0 &&
+                 (int)difftime(now, w->s->first_error_time) >= p->error_escalation_time &&
                  w->s->state != JK_LB_STATE_RECOVER) {
             if (JK_IS_DEBUG_LEVEL(l))
                 jk_log(l, JK_LOG_DEBUG,
@@ -704,24 +728,15 @@ static jk_uint64_t decay_load(lb_worker_t *p,
     return curmax;
 }
 
-static int JK_METHOD maintain_workers(jk_worker_t *p, time_t now, jk_logger_t *l)
+static int JK_METHOD maintain_workers(jk_worker_t *p, time_t now, int global, jk_logger_t *l)
 {
     unsigned int i = 0;
     jk_uint64_t curmax = 0;
-    long delta;
 
     JK_TRACE_ENTER(l);
     if (p && p->worker_private) {
         lb_worker_t *lb = (lb_worker_t *)p->worker_private;
 
-        for (i = 0; i < lb->num_of_workers; i++) {
-            if (lb->lb_workers[i].worker->maintain) {
-                lb->lb_workers[i].worker->maintain(lb->lb_workers[i].worker, now, l);
-            }
-        }
-
-        jk_shm_lock();
-
         /* Now we check for global maintenance (once for all processes).
          * Checking workers for recovery and applying decay to the
          * load values should not be done by each process individually.
@@ -729,20 +744,65 @@ static int JK_METHOD maintain_workers(jk_worker_t *p, time_t now, jk_logger_t *l
          * Since it's possible that we come here a few milliseconds
          * before the interval has passed, we allow a little tolerance.
          */
-        delta = (long)difftime(now, lb->s->last_maintain_time) + JK_LB_MAINTAIN_TOLERANCE;
-        if (delta >= lb->maintain_time) {
+        if (global == JK_TRUE) {
+            time_t exponent = JK_LB_DECAY_MULT * difftime(now, lb->s->last_maintain_time) / lb->maintain_time;
             lb->s->last_maintain_time = now;
+
             if (JK_IS_DEBUG_LEVEL(l))
                 jk_log(l, JK_LOG_DEBUG,
                        "decay with 2^%d",
-                       JK_LB_DECAY_MULT * delta / lb->maintain_time);
-            curmax = decay_load(lb, JK_LB_DECAY_MULT * delta / lb->maintain_time, l);
+                       exponent);
+
+            jk_shm_lock();
+
+            curmax = decay_load(lb, exponent, l);
+
             if (!recover_workers(lb, curmax, now, l)) {
                 force_recovery(lb, NULL, l);
             }
+
+            /*
+             * Checking workers for idleness.
+             */
+            for (i = 0; i < lb->num_of_workers; i++) {
+                ajp_worker_t *aw = lb->lb_workers[i].worker->worker_private;
+                if (aw->s->state == JK_AJP_STATE_OK &&
+                    aw->s->used == aw->s->used_snapshot)
+                    aw->s->state = JK_AJP_STATE_IDLE;
+                aw->s->used_snapshot = aw->s->used;
+            }
+
+            jk_shm_unlock();
         }
 
-        jk_shm_unlock();
+        for (i = 0; i < lb->num_of_workers; i++) {
+            if (lb->lb_workers[i].worker->maintain) {
+                lb->lb_workers[i].worker->maintain(lb->lb_workers[i].worker, now, global, l);
+            }
+        }
+
+    }
+    else {
+        JK_LOG_NULL_PARAMS(l);
+    }
+
+    JK_TRACE_EXIT(l);
+    return JK_TRUE;
+}
+
+static int JK_METHOD shutdown_workers(jk_worker_t *p, jk_logger_t *l)
+{
+    unsigned int i = 0;
+
+    JK_TRACE_ENTER(l);
+    if (p && p->worker_private) {
+        lb_worker_t *lb = (lb_worker_t *)p->worker_private;
+
+        for (i = 0; i < lb->num_of_workers; i++) {
+            if (lb->lb_workers[i].worker->shutdown) {
+                lb->lb_workers[i].worker->shutdown(lb->lb_workers[i].worker, l);
+            }
+        }
 
     }
     else {
@@ -881,6 +941,8 @@ static int find_bysession_route(jk_ws_service_t *s,
     if (candidate < 0) {
         uses_domain = 1;
         candidate = find_best_bydomain(s, p, session_route, states, l);
+    } else {
+        s->sticky = JK_TRUE;
     }
     if (candidate >= 0) {
         lb_sub_worker_t wr = p->lb_workers[candidate];
@@ -943,8 +1005,10 @@ static int find_failover_worker(jk_ws_service_t *s,
             break;
         }
     }
-    if (redirect)
+    if (redirect) {
         rc = find_bysession_route(s, p, redirect, states, l);
+        s->sticky = JK_FALSE;
+    }
     return rc;
 }
 
@@ -971,6 +1035,7 @@ static int get_most_suitable_worker(jk_ws_service_t *s,
     int rc = -1;
 
     JK_TRACE_ENTER(l);
+    s->sticky = JK_FALSE;
     if (p->num_of_workers == 1) {
         /* No need to find the best worker
          * if there is a single one
@@ -982,6 +1047,7 @@ static int get_most_suitable_worker(jk_ws_service_t *s,
             activation = p->lb_workers[0].activation;
         if (JK_WORKER_USABLE_STICKY(states[0], activation)) {
             if (activation != JK_LB_ACTIVATION_DISABLED) {
+                s->sticky = JK_TRUE;
                 JK_TRACE_EXIT(l);
                 return 0;
             }
@@ -1082,6 +1148,7 @@ static void lb_add_log_items(jk_ws_service_t *s,
                              lb_sub_worker_t *w,
                              jk_logger_t *l)
 {
+    /* ADJUST JK_LB_NOTES_COUNT WHEN ADDING MORE NOTES HERE! */
     ajp_worker_t *aw = (ajp_worker_t *)w->worker->worker_private;
     const char **log_values = jk_pool_alloc(s->pool, sizeof(char *) * JK_LB_NOTES_COUNT);
     char *buf = jk_pool_alloc(s->pool, sizeof(char *) * JK_LB_NOTES_COUNT * JK_LB_UINT64_STR_SZ);
@@ -1159,11 +1226,13 @@ static int JK_METHOD service(jk_endpoint_t *e,
         jk_lb_pull(p->worker, JK_FALSE, l);
     for (i = 0; i < num_of_workers; i++) {
         lb_sub_worker_t *rec = &(p->worker->lb_workers[i]);
+        ajp_worker_t *aw = (ajp_worker_t *)rec->worker->worker_private;
         if (rec->s->state == JK_LB_STATE_BUSY) {
-            if (ajp_has_endpoint(rec->worker, l)) {
+            if ((aw->busy_limit <= 0 || aw->s->busy < aw->busy_limit) &&
+                ajp_has_endpoint(rec->worker, l)) {
                 if (JK_IS_DEBUG_LEVEL(l))
                     jk_log(l, JK_LOG_DEBUG,
-                           "worker %s busy count fixed",
+                           "worker %s busy state ended",
                            rec->name);
                 rec->s->state = JK_LB_STATE_OK;
             }
@@ -1244,8 +1313,8 @@ static int JK_METHOD service(jk_endpoint_t *e,
 
             if (JK_IS_DEBUG_LEVEL(l))
                 jk_log(l, JK_LOG_DEBUG,
-                       "service worker=%s route=%s",
-                       rec->name, s->route);
+                       "service worker=%s route=%s failover=%s",
+                       rec->name, s->route, s->sticky ? "false" : "true");
 
             if (p->worker->lblock == JK_LB_LOCK_PESSIMISTIC)
                 jk_shm_lock();
@@ -1279,6 +1348,7 @@ static int JK_METHOD service(jk_endpoint_t *e,
                 int service_stat = JK_UNSET;
                 jk_uint64_t rd = 0;
                 jk_uint64_t wr = 0;
+                int busy;
                 /* Reset endpoint read and write sizes for
                  * this request.
                  */
@@ -1288,10 +1358,9 @@ static int JK_METHOD service(jk_endpoint_t *e,
                     jk_shm_lock();
 
                 /* Increment the number of workers serving request */
-                p->worker->s->busy++;
-                rec->s->busy++;
-                if (p->worker->s->busy > p->worker->s->max_busy)
-                    p->worker->s->max_busy = p->worker->s->busy;
+                busy = JK_ATOMIC_INCREMENT(&(p->worker->s->busy));
+                if (busy > p->worker->s->max_busy)
+                    p->worker->s->max_busy = busy;
                 if (p->worker->lbmethod == JK_LB_METHOD_REQUESTS ||
                     p->worker->lbmethod == JK_LB_METHOD_BUSYNESS ||
                     (!sessionid &&
@@ -1305,6 +1374,41 @@ static int JK_METHOD service(jk_endpoint_t *e,
                 if (p->worker->lblock == JK_LB_LOCK_PESSIMISTIC)
                     jk_shm_unlock();
 
+                if (!s->sticky && (s->extension.set_session_cookie || p->worker->set_session_cookie)) {
+                    char **old_names = s->resp_headers_names;
+                    char **old_values = s->resp_headers_values;
+                    s->resp_headers_names  = jk_pool_alloc(s->pool,
+                                                      (s->num_resp_headers + 1) * sizeof(char *));
+                    s->resp_headers_values = jk_pool_alloc(s->pool,
+                                                      (s->num_resp_headers + 1) * sizeof(char *));
+                    if (!s->resp_headers_names || !s->resp_headers_values) {
+                        jk_log(l, JK_LOG_ERROR,
+                            "Failed allocating %d new response headers.", s->num_resp_headers + 1);
+                        s->resp_headers_names = old_names;
+                        s->resp_headers_values = old_values;
+                    } else if (s->num_resp_headers) {
+                        memcpy(s->resp_headers_names, old_names, s->num_resp_headers * sizeof(char *));
+                        memcpy(s->resp_headers_values, old_values, s->num_resp_headers * sizeof(char *));
+                    }
+                    s->resp_headers_names[s->num_resp_headers] = "Set-Cookie";
+                    s->resp_headers_values[s->num_resp_headers] = jk_pool_strcatv(s->pool, p->worker->session_cookie,
+                                                                                  "=.", rec->route,
+                                                                                  NULL);
+                    if (p->worker->session_cookie_path && *p->worker->session_cookie_path) {
+                        s->resp_headers_values[s->num_resp_headers] = jk_pool_strcatv(s->pool, s->resp_headers_values[s->num_resp_headers],
+                                                                                      ";PATH=", p->worker->session_cookie_path,
+                                                                                      NULL);
+                    }
+                    s->resp_headers_values[s->num_resp_headers] = jk_pool_strcatv(s->pool, s->resp_headers_values[s->num_resp_headers],
+                                                                                  ";HttpOnly",
+                                                                                  (s->is_ssl ? ";Secure" : ""),
+                                                                                  NULL);
+                    if (JK_IS_DEBUG_LEVEL(l))
+                        jk_log(l, JK_LOG_DEBUG, "Added cookie header '%s' with value '%s' ",
+                               s->resp_headers_names[s->num_resp_headers],
+                               s->resp_headers_values[s->num_resp_headers]);
+                    s->num_resp_headers++;
+                }
                 service_stat = end->service(end, s, l, &is_service_error);
                 rd = end->rd;
                 wr = end->wr;
@@ -1353,17 +1457,15 @@ static int JK_METHOD service(jk_endpoint_t *e,
                  * Check if the busy was reset to zero by graceful
                  * restart of the server.
                  */
-                if (p->worker->s->busy)
-                    p->worker->s->busy--;
-                if (rec->s->busy)
-                    rec->s->busy--;
+                JK_ATOMIC_DECREMENT(&(p->worker->s->busy));
                 if (service_stat == JK_TRUE) {
                     /*
                      * Successful request.
                      */
                     rec->s->state  = JK_LB_STATE_OK;
                     p->states[rec->i] = JK_LB_STATE_OK;
-                    rec->s->error_time = 0;
+                    rec->s->first_error_time = 0;
+                    rec->s->last_error_time = 0;
                     rc = JK_TRUE;
                     recoverable = JK_UNSET;
                 }
@@ -1374,7 +1476,8 @@ static int JK_METHOD service(jk_endpoint_t *e,
                      */
                     rec->s->state  = JK_LB_STATE_OK;
                     p->states[rec->i] = JK_LB_STATE_ERROR;
-                    rec->s->error_time = 0;
+                    rec->s->first_error_time = 0;
+                    rec->s->last_error_time = 0;
                     rc = JK_CLIENT_ERROR;
                     recoverable = JK_FALSE;
                 }
@@ -1407,7 +1510,18 @@ static int JK_METHOD service(jk_endpoint_t *e,
                      */
                     rec->s->state  = JK_LB_STATE_OK;
                     p->states[rec->i] = JK_LB_STATE_ERROR;
-                    rec->s->error_time = 0;
+                    rec->s->first_error_time = 0;
+                    rec->s->last_error_time = 0;
+                    rc = JK_FALSE;
+                }
+                else if (service_stat == JK_BUSY_ERROR) {
+                    /*
+                     * Node was busy.
+                     * Do not try to reuse the same node for the same request.
+                     * Failing over to another node could help.
+                     */
+                    rec->s->state  = JK_LB_STATE_BUSY;
+                    p->states[rec->i] = JK_LB_STATE_BUSY;
                     rc = JK_FALSE;
                 }
                 else if (service_stat == JK_STATUS_FATAL_ERROR) {
@@ -1420,7 +1534,8 @@ static int JK_METHOD service(jk_endpoint_t *e,
                     rec->s->errors++;
                     rec->s->state = JK_LB_STATE_ERROR;
                     p->states[rec->i] = JK_LB_STATE_ERROR;
-                    rec->s->error_time = time(NULL);
+                    rec->s->first_error_time = time(NULL);
+                    rec->s->last_error_time = rec->s->first_error_time;
                     rc = JK_FALSE;
                 }
                 else if (service_stat == JK_REPLY_TIMEOUT) {
@@ -1434,7 +1549,8 @@ static int JK_METHOD service(jk_endpoint_t *e,
                         rec->s->errors++;
                         rec->s->state = JK_LB_STATE_ERROR;
                         p->states[rec->i] = JK_LB_STATE_ERROR;
-                        rec->s->error_time = time(NULL);
+                        rec->s->first_error_time = time(NULL);
+                        rec->s->last_error_time = rec->s->first_error_time;
                     }
                     else {
                         /*
@@ -1456,10 +1572,10 @@ static int JK_METHOD service(jk_endpoint_t *e,
                      */
                     time_t now = time(NULL);
                     rec->s->errors++;
-                    if (rec->s->busy == 0 ||
+                    if (aw->s->busy == 0 ||
                         p->worker->error_escalation_time == 0 ||
-                        (rec->s->error_time > 0 &&
-                         (int)difftime(now, rec->s->error_time) >= p->worker->error_escalation_time)) {
+                        (rec->s->first_error_time > 0 &&
+                         (int)difftime(now, rec->s->first_error_time) >= p->worker->error_escalation_time)) {
                         if (JK_IS_DEBUG_LEVEL(l))
                             jk_log(l, JK_LOG_DEBUG,
                                    "worker %s escalating local error to global error",
@@ -1467,9 +1583,10 @@ static int JK_METHOD service(jk_endpoint_t *e,
                         rec->s->state = JK_LB_STATE_ERROR;
                     }
                     p->states[rec->i] = JK_LB_STATE_ERROR;
-                    if (rec->s->error_time == 0) {
-                        rec->s->error_time = now;
+                    if (rec->s->first_error_time == 0) {
+                        rec->s->first_error_time = now;
                     }
+                    rec->s->last_error_time = now;
                     rc = JK_FALSE;
                 }
                 if (p->worker->lblock == JK_LB_LOCK_PESSIMISTIC)
@@ -1621,7 +1738,11 @@ static int JK_METHOD validate(jk_worker_t *pThis,
             }
             memset(p->lb_workers, 0, num_of_workers * sizeof(lb_sub_worker_t));
             for (i = 0; i < num_of_workers; i++) {
-                p->lb_workers[i].s = jk_shm_alloc_lb_sub_worker(&p->p, p->s->h.id, worker_names[i]);
+                if (jk_check_attribute_length("host name", worker_names[i], l) == JK_FALSE) {
+                    JK_TRACE_EXIT(l);
+                    return JK_FALSE;
+                }
+                p->lb_workers[i].s = jk_shm_alloc_lb_sub_worker(&p->p, p->s->h.id, worker_names[i], l);
                 if (p->lb_workers[i].s == NULL) {
                     jk_log(l, JK_LOG_ERROR,
                            "allocating lb sub worker record from shared memory");
@@ -1667,18 +1788,34 @@ static int JK_METHOD validate(jk_worker_t *pThis,
                     p->max_packet_size = ms;
                 p->lb_workers[i].distance =
                     jk_get_distance(props, worker_names[i]);
-                if ((s = jk_get_worker_route(props, worker_names[i], NULL)))
+                if ((s = jk_get_worker_route(props, worker_names[i], NULL))) {
+                    if (jk_check_attribute_length("route", s, l) == JK_FALSE) {
+                        JK_TRACE_EXIT(l);
+                        return JK_FALSE;
+                    }
                     strncpy(p->lb_workers[i].route, s, JK_SHM_STR_SIZ);
+                }
                 else
                     strncpy(p->lb_workers[i].route, worker_names[i], JK_SHM_STR_SIZ);
-                if ((s = jk_get_worker_domain(props, worker_names[i], NULL)))
+                if ((s = jk_get_worker_domain(props, worker_names[i], NULL))) {
+                    if (jk_check_attribute_length("domain", s, l) == JK_FALSE) {
+                        JK_TRACE_EXIT(l);
+                        return JK_FALSE;
+                    }
                     strncpy(p->lb_workers[i].domain, s, JK_SHM_STR_SIZ);
-                if ((s = jk_get_worker_redirect(props, worker_names[i], NULL)))
+                }
+                if ((s = jk_get_worker_redirect(props, worker_names[i], NULL))) {
+                    if (jk_check_attribute_length("redirect", s, l) == JK_FALSE) {
+                        JK_TRACE_EXIT(l);
+                        return JK_FALSE;
+                    }
                     strncpy(p->lb_workers[i].redirect, s, JK_SHM_STR_SIZ);
+                }
 
                 p->lb_workers[i].s->lb_value = 0;
                 p->lb_workers[i].s->state = JK_LB_STATE_IDLE;
-                p->lb_workers[i].s->error_time = 0;
+                p->lb_workers[i].s->first_error_time = 0;
+                p->lb_workers[i].s->last_error_time = 0;
                 p->lb_workers[i].s->elected_snapshot = 0;
                 p->lb_workers[i].s->sessions = 0;
                 p->lb_workers[i].activation =
@@ -1723,7 +1860,7 @@ static int JK_METHOD validate(jk_worker_t *pThis,
                     }
                     if (JK_IS_DEBUG_LEVEL(l)) {
                         jk_log(l, JK_LOG_DEBUG,
-                               "Balanced worker %i has name %s and route %s in domain %s",
+                               "Balanced worker %d has name %s and route %s in domain %s",
                                i,
                                p->lb_workers[i].name,
                                p->lb_workers[i].route,
@@ -1736,7 +1873,7 @@ static int JK_METHOD validate(jk_worker_t *pThis,
                     for (j = 0; j < i; j++) {
                         if (strcmp(p->lb_workers[i].route, p->lb_workers[j].route) == 0) {
                             jk_log(l, JK_LOG_ERROR,
-                                   "Balanced workers number %i (%s) and %i (%s) share the same route %s - aborting configuration!",
+                                   "Balanced workers number %d (%s) and %d (%s) share the same route %s - aborting configuration!",
                                    i,
                                    p->lb_workers[i].name,
                                    j,
@@ -1763,6 +1900,7 @@ static int JK_METHOD init(jk_worker_t *pThis,
                           jk_worker_env_t *we, jk_logger_t *log)
 {
     int i;
+    const char *s;
 
     lb_worker_t *p = (lb_worker_t *)pThis->worker_private;
     JK_TRACE_ENTER(log);
@@ -1788,15 +1926,32 @@ static int JK_METHOD init(jk_worker_t *pThis,
     p->s->last_reset = p->s->last_maintain_time;
 
     p->lbmethod = jk_get_lb_method(props, p->name);
+#ifdef JK_ATOMIC_MISSING
+    if (p->lbmethod == JK_LB_METHOD_BUSYNESS) {
+        jk_log(log, JK_LOG_WARNING, "Missing support for atomics: "
+               "LB method 'busyness' not recommended");
+    }
+#endif
     p->lblock   = jk_get_lb_lock(props, p->name);
-    strncpy(p->session_cookie,
-            jk_get_lb_session_cookie(props, p->name, JK_SESSION_IDENTIFIER),
-            JK_SHM_STR_SIZ);
-    strncpy(p->session_path,
-            jk_get_lb_session_path(props, p->name, JK_PATH_SESSION_IDENTIFIER),
-            JK_SHM_STR_SIZ);
-    strcpy(p->s->session_cookie, p->session_cookie);
-    strcpy(p->s->session_path, p->session_path);
+    s = jk_get_lb_session_cookie(props, p->name, JK_SESSION_IDENTIFIER);
+    if (jk_check_attribute_length("session_cookie", s, log) == JK_FALSE) {
+        JK_TRACE_EXIT(log);
+        return JK_FALSE;
+    }
+    strncpy(p->session_cookie, s, JK_SHM_STR_SIZ);
+    s = jk_get_lb_session_path(props, p->name, JK_PATH_SESSION_IDENTIFIER);
+    if (jk_check_attribute_length("session_path", s, log) == JK_FALSE) {
+        JK_TRACE_EXIT(log);
+        return JK_FALSE;
+    }
+    strncpy(p->session_path, s, JK_SHM_STR_SIZ);
+    p->set_session_cookie = jk_get_lb_set_session_cookie(props, p->name, JK_FALSE);
+    s = jk_get_lb_session_cookie_path(props, p->name, "/");
+    if (jk_check_attribute_length("session_cookie_path", s, log) == JK_FALSE) {
+        JK_TRACE_EXIT(log);
+        return JK_FALSE;
+    }
+    strncpy(p->session_cookie_path, s, JK_SHM_STR_SIZ);
 
     JK_INIT_CS(&(p->cs), i);
     if (i == JK_FALSE) {
@@ -1887,7 +2042,7 @@ int JK_METHOD lb_worker_factory(jk_worker_t **w,
                         private_data->buf,
                         sizeof(jk_pool_atom_t) * TINY_POOL_SIZE);
 
-        private_data->s = jk_shm_alloc_lb_worker(&private_data->p, name);
+        private_data->s = jk_shm_alloc_lb_worker(&private_data->p, name, l);
         if (!private_data->s) {
             free(private_data);
             JK_TRACE_EXIT(l);
@@ -1902,10 +2057,11 @@ int JK_METHOD lb_worker_factory(jk_worker_t **w,
         private_data->worker.get_endpoint = get_endpoint;
         private_data->worker.destroy = destroy;
         private_data->worker.maintain = maintain_workers;
+        private_data->worker.shutdown = shutdown_workers;
         private_data->recover_wait_time = WAIT_BEFORE_RECOVER;
         private_data->error_escalation_time = private_data->recover_wait_time / 2;
         private_data->max_reply_timeouts = 0;
-        private_data->max_packet_size = DEF_BUFFER_SZ;
+        private_data->max_packet_size = AJP13_DEF_PACKET_SIZE;
         private_data->sequence = 0;
         private_data->next_offset = 0;
         *w = &private_data->worker;
