@@ -19,7 +19,7 @@
  * Description: General purpose map object                                 *
  * Author:      Gal Shachor <shachor@il.ibm.com>                           *
  * Author:      Mladen Turk <mturk@apache.org>                             *
- * Version:     $Revision: 1306287 $                                          *
+ * Version:     $Revision: 1647863 $                                          *
  ***************************************************************************/
 #if defined(AS400) && !defined(AS400_UTF8)
 #include "apr_xlate.h"
@@ -36,32 +36,51 @@
 #define JK_MAP_REFERENCE    (".reference")
 #define JK_MAP_REFERENCE_SZ (strlen(JK_MAP_REFERENCE))
 
-/* Compute the "checksum" for a key, consisting of the first
- * 4 bytes, packed into an int.
- * This checksum allows us to do a single integer
- * comparison as a fast check to determine whether we can
- * skip a strcmp
+/* Taken from APR tables/apr_hash.c */
+/*
+ * This is the popular `times 33' hash algorithm which is used by
+ * perl and also appears in Berkeley DB. This is one of the best
+ * known hash functions for strings because it is both computed
+ * very fast and distributes very well.
+ *
+ * The originator may be Dan Bernstein but the code in Berkeley DB
+ * cites Chris Torek as the source. The best citation I have found
+ * is "Chris Torek, Hash function for text in C, Usenet message
+ * <27038@mimsy.umd.edu> in comp.lang.c , October, 1990." in Rich
+ * Salz's USENIX 1992 paper about INN which can be found at
+ * <http://citeseer.nj.nec.com/salz92internetnews.html>.
+ *
+ * The magic of number 33, i.e. why it works better than many other
+ * constants, prime or not, has never been adequately explained by
+ * anyone. So I try an explanation: if one experimentally tests all
+ * multipliers between 1 and 256 (as I did while writing a low-level
+ * data structure library some time ago) one detects that even
+ * numbers are not useable at all. The remaining 128 odd numbers
+ * (except for the number 1) work more or less all equally well.
+ * They all distribute in an acceptable way and this way fill a hash
+ * table with an average percent of approx. 86%.
+ *
+ * If one compares the chi^2 values of the variants (see
+ * Bob Jenkins ``Hashing Frequently Asked Questions'' at
+ * http://burtleburtle.net/bob/hash/hashfaq.html for a description
+ * of chi^2), the number 33 not even has the best value. But the
+ * number 33 and a few other equally good numbers like 17, 31, 63,
+ * 127 and 129 have nevertheless a great advantage to the remaining
+ * numbers in the large set of possible multipliers: their multiply
+ * operation can be replaced by a faster operation based on just one
+ * shift plus either a single addition or subtraction operation. And
+ * because a hash function has to both distribute good _and_ has to
+ * be very fast to compute, those few numbers should be preferred.
+ *
+ *                  -- Ralf S. Engelschall <rse@engelschall.com>
  */
-#define COMPUTE_KEY_CHECKSUM(key, checksum)    \
-{                                              \
-    const char *k = (key);                     \
-    unsigned int c = (unsigned int)*k;         \
-    (checksum) = c;                            \
-    (checksum) <<= 8;                          \
-    if (c) {                                   \
-        c = (unsigned int)*++k;                \
-        checksum |= c;                         \
-    }                                          \
-    (checksum) <<= 8;                          \
-    if (c) {                                   \
-        c = (unsigned int)*++k;                \
-        checksum |= c;                         \
-    }                                          \
-    (checksum) <<= 8;                          \
-    if (c) {                                   \
-        c = (unsigned int)*++k;                \
-        checksum |= c;                         \
-    }                                          \
+#define COMPUTE_KEY_HASH(key, hash)    \
+{                                                    \
+    const unsigned char *p;                          \
+    hash = 0;                                        \
+    for (p = (const unsigned char *)key; *p; p++) {  \
+        hash = hash * 33 + *p;                       \
+    }                                                \
 }
 
 static volatile int global_map_id = 0;
@@ -130,10 +149,10 @@ void *jk_map_get(jk_map_t *m, const char *name, const void *def)
 
     if (m && name) {
         unsigned int i;
-        unsigned int key;
-        COMPUTE_KEY_CHECKSUM(name, key)
+        unsigned int hash;
+        COMPUTE_KEY_HASH(name, hash)
         for (i = 0; i < m->size; i++) {
-            if (m->keys[i] == key && strcmp(m->names[i], name) == 0) {
+            if (m->keys[i] == hash && strcmp(m->names[i], name) == 0) {
                 rc = m->values[i];
                 break;
             }
@@ -148,10 +167,10 @@ int jk_map_get_id(jk_map_t *m, const char *name)
     int rc = -1;
     if (m && name) {
         unsigned int i;
-        unsigned int key;
-        COMPUTE_KEY_CHECKSUM(name, key)
+        unsigned int hash;
+        COMPUTE_KEY_HASH(name, hash)
         for (i = 0; i < m->size; i++) {
-            if (m->keys[i] == key && strcmp(m->names[i], name) == 0) {
+            if (m->keys[i] == hash && strcmp(m->names[i], name) == 0) {
                 rc = i;
                 break;
             }
@@ -167,10 +186,10 @@ const char *jk_map_get_string(jk_map_t *m, const char *name, const char *def)
 
     if (m && name) {
         unsigned int i;
-        unsigned int key;
-        COMPUTE_KEY_CHECKSUM(name, key)
+        unsigned int hash;
+        COMPUTE_KEY_HASH(name, hash)
         for (i = 0; i < m->size; i++) {
-            if (m->keys[i] == key && strcmp(m->names[i], name) == 0) {
+            if (m->keys[i] == hash && strcmp(m->names[i], name) == 0) {
                 rc = m->values[i];
                 break;
             }
@@ -187,29 +206,27 @@ int jk_map_get_int(jk_map_t *m, const char *name, int def)
     const char *rc;
     size_t len;
     int int_res;
-    int multit = 1;
 
     sprintf(buf, "%d", def);
     rc = jk_map_get_string(m, name, buf);
 
     len = strlen(rc);
     if (len) {
-        char *lastchar = &buf[0] + len - 1;
-        strcpy(buf, rc);
+        const char *lastchar = &rc[0] + len - 1;
+        int multit = 1;
         if ('m' == *lastchar || 'M' == *lastchar) {
-            *lastchar = '\0';
             multit = 1024 * 1024;
         }
         else if ('k' == *lastchar || 'K' == *lastchar) {
-            *lastchar = '\0';
             multit = 1024;
         }
-        int_res = atoi(buf);
+        /* Safe because atoi() will stop at any non-numeric lastchar */
+        int_res = atoi(rc) * multit;
     }
     else
         int_res = def;
 
-    return int_res * multit;
+    return int_res;
 }
 
 double jk_map_get_double(jk_map_t *m, const char *name, double def)
@@ -269,13 +286,12 @@ char **jk_map_get_string_list(jk_map_t *m,
 #endif
 
         {
-
             if (idex == capacity) {
                 ar = jk_pool_realloc(&m->p,
                                      sizeof(char *) * (capacity + 5),
                                      ar, sizeof(char *) * capacity);
                 if (!ar) {
-                    return JK_FALSE;
+                    return NULL;
                 }
                 capacity += 5;
             }
@@ -289,29 +305,26 @@ char **jk_map_get_string_list(jk_map_t *m,
     return ar;
 }
 
-int jk_map_get_int_list(jk_map_t *m,
-                        const char *name,
-                        int *list,
-                        unsigned int list_len,
-                        const char *def)
+int *jk_map_get_int_list(jk_map_t *m,
+                         const char *name,
+                         unsigned int *list_len,
+                         const char *def)
 {
     const char *l = jk_map_get_string(m, name, def);
+    int *ar = NULL;
 
 #ifdef _MT_CODE_PTHREAD
     char *lasts;
 #endif
 
-    if (!list_len)
-        return 0;
-
     if (l) {
-        unsigned int capacity = list_len;
-        unsigned int index = 0;
+        unsigned int capacity = 0;
+        unsigned int idex = 0;
         char *p;
         char *v = jk_pool_strdup(&m->p, l);
 
         if (!v) {
-            return 0;
+            return NULL;
         }
 
         /*
@@ -326,16 +339,22 @@ int jk_map_get_int_list(jk_map_t *m,
 #endif
 
         {
-            if (index < capacity) {
-                list[index] = atoi(p);
-                index++;
+            if (idex == capacity) {
+                ar = jk_pool_realloc(&m->p,
+                                     sizeof(int) * (capacity + 5),
+                                     ar, sizeof(int) * capacity);
+                if (!ar) {
+                    return NULL;
+                }
+                capacity += 5;
             }
-            else
-                break;
+            ar[idex] = atoi(p);
+            idex++;
         }
-        return index;
+
+        *list_len = idex;
     }
-    return 0;
+    return ar;
 }
 
 int jk_map_add(jk_map_t *m, const char *name, const void *value)
@@ -343,14 +362,14 @@ int jk_map_add(jk_map_t *m, const char *name, const void *value)
     int rc = JK_FALSE;
 
     if (m && name) {
-        unsigned int key;
-        COMPUTE_KEY_CHECKSUM(name, key)
+        unsigned int hash;
+        COMPUTE_KEY_HASH(name, hash)
         map_realloc(m);
 
         if (m->size < m->capacity) {
             m->values[m->size] = value;
             m->names[m->size] = jk_pool_strdup(&m->p, name);
-            m->keys[m->size] = key;
+            m->keys[m->size] = hash;
             m->size++;
             rc = JK_TRUE;
         }
@@ -365,10 +384,10 @@ int jk_map_put(jk_map_t *m, const char *name, const void *value, void **old)
 
     if (m && name) {
         unsigned int i;
-        unsigned int key;
-        COMPUTE_KEY_CHECKSUM(name, key)
+        unsigned int hash;
+        COMPUTE_KEY_HASH(name, hash)
         for (i = 0; i < m->size; i++) {
-            if (m->keys[i] == key && strcmp(m->names[i], name) == 0) {
+            if (m->keys[i] == hash && strcmp(m->names[i], name) == 0) {
                 break;
             }
         }
@@ -454,8 +473,8 @@ int jk_map_read_property(jk_map_t *m, jk_map_t *env, const char *str,
     char *prp = &buf[0];
 
     if (strlen(str) > LENGTH_OF_LINE) {
-        jk_log(l, JK_LOG_WARNING,
-               "Line to long (%d > %d), ignoring entry",
+        jk_log(l, JK_LOG_ERROR,
+               "Line too long (%d > %d), ignoring entry",
                strlen(str), LENGTH_OF_LINE);
         return JK_FALSE;
     }
