@@ -19,7 +19,7 @@
  * Description: Shared Memory support                                      *
  * Author:      Mladen Turk <mturk@jboss.com>                              *
  * Author:      Rainer Jung <rjung@apache.org>                             *
- * Version:     $Revision: 1334392 $                                        *
+ * Version:     $Revision: 1665454 $                                        *
  ***************************************************************************/
 
 #include "jk_global.h"
@@ -42,7 +42,8 @@ struct jk_shm_header_data
     unsigned int pos;
     unsigned int childs;
     unsigned int workers;
-    time_t       modified;
+    unsigned int maintain_checking;
+    volatile time_t maintain_time;
 };
 
 typedef struct jk_shm_header_data jk_shm_header_data_t;
@@ -81,18 +82,47 @@ typedef struct jk_shm jk_shm_t;
 
 static const char shm_signature[] = { JK_SHM_MAGIC };
 static jk_shm_t jk_shmem = { 0, 0, 0, 0, NULL, NULL, -1, -1, 0, NULL};
-static time_t jk_workers_modified_time = 0;
-static time_t jk_workers_access_time = 0;
 #if defined (WIN32)
 static HANDLE jk_shm_map = NULL;
 static HANDLE jk_shm_hlock = NULL;
 #endif
 static int jk_shm_inited_cs = 0;
 
+static size_t jk_shm_calculate_slot_size()
+{
+    int align = 64;
+    size_t slot_size = 0;
+    size_t this_size = 0;
+
+    this_size = sizeof(jk_shm_header_data_t);
+    if (this_size > slot_size)
+        slot_size = this_size;
+
+    this_size = sizeof(jk_shm_worker_header_t);
+    if (this_size > slot_size)
+        slot_size = this_size;
+
+    this_size = sizeof(jk_shm_ajp_worker_t);
+    if (this_size > slot_size)
+        slot_size = this_size;
+
+    this_size = sizeof(jk_shm_lb_sub_worker_t);
+    if (this_size > slot_size)
+        slot_size = this_size;
+
+    this_size = sizeof(jk_shm_lb_worker_t);
+    if (this_size > slot_size)
+        slot_size = this_size;
+
+    slot_size = JK_ALIGN(slot_size, align);
+    return slot_size;
+}
+
 /* Calculate needed shm size */
 int jk_shm_calculate_size(jk_map_t *init_data, jk_logger_t *l)
 {
     char **worker_list;
+    size_t needed_slot_size = 0;
     unsigned int i;
     unsigned int num_of_workers;
     int num_of_ajp_workers = 0;
@@ -108,6 +138,17 @@ int jk_shm_calculate_size(jk_map_t *init_data, jk_logger_t *l)
         JK_TRACE_EXIT(l);
         return 0;
     }
+    needed_slot_size = jk_shm_calculate_slot_size();
+    if (JK_IS_DEBUG_LEVEL(l))
+        jk_log(l, JK_LOG_DEBUG, "JK_SHM_SLOT_SIZE defined as %d, need at least %d",
+               JK_SHM_SLOT_SIZE , needed_slot_size);
+    if (needed_slot_size > JK_SHM_SLOT_SIZE) {
+        jk_log(l, JK_LOG_ERROR,
+               "JK_SHM_SLOT_SIZE %d is too small, need at least %d",
+               JK_SHM_SLOT_SIZE, needed_slot_size);
+        JK_TRACE_EXIT(l);
+        return 0;
+    }
     for (i = 0; i < num_of_workers; i++) {
         const char *type = jk_get_worker_type(init_data, worker_list[i]);
 
@@ -285,6 +326,8 @@ int jk_shm_open(const char *fname, int sz, jk_logger_t *l)
                JK_SHM_MAGIC_SIZ);
         jk_shmem.hdr->h.data.size = sz;
         jk_shmem.hdr->h.data.childs = 1;
+        jk_shmem.hdr->h.data.maintain_checking = 0;
+        jk_shmem.hdr->h.data.maintain_time = time(NULL);
     }
     else {
         jk_shmem.hdr->h.data.childs++;
@@ -620,6 +663,8 @@ static int do_shm_open(const char *fname, int attached,
         memcpy(jk_shmem.hdr->h.data.magic, shm_signature, JK_SHM_MAGIC_SIZ);
         jk_shmem.hdr->h.data.size = sz;
         jk_shmem.hdr->h.data.childs = 1;
+        jk_shmem.hdr->h.data.maintain_checking = 0;
+        jk_shmem.hdr->h.data.maintain_time = time(NULL);
         if (JK_IS_DEBUG_LEVEL(l))
             jk_log(l, JK_LOG_DEBUG,
                    "Initialized shared memory %s size=%u free=%u addr=%#lx",
@@ -762,30 +807,17 @@ void jk_shm_close(jk_logger_t *l)
 
 #endif
 
-void *jk_shm_alloc(jk_pool_t *p)
-{
-    void *rc = NULL;
-
-    if (jk_shmem.hdr) {
-        jk_shm_lock();
-        if ((jk_shmem.hdr->h.data.size - jk_shmem.hdr->h.data.pos) >= JK_SHM_SLOT_SIZE) {
-            rc = &(jk_shmem.hdr->buf[jk_shmem.hdr->h.data.pos]);
-            jk_shmem.hdr->h.data.pos += JK_SHM_SLOT_SIZE;
-        }
-        jk_shm_unlock();
-    }
-    else if (p)
-        rc = jk_pool_alloc(p, JK_SHM_SLOT_SIZE);
-
-    return rc;
-}
-
 jk_shm_worker_header_t *jk_shm_alloc_worker(jk_pool_t *p, int type,
-                                            int parent_id, const char *name)
+                                            int parent_id, const char *name,
+                                            jk_logger_t *l)
 {
     unsigned int i;
     jk_shm_worker_header_t *w = 0;
 
+    if (jk_check_attribute_length("name", name, l) == JK_FALSE) {
+        return NULL;
+    }
+
     if (jk_shmem.hdr) {
         jk_shm_lock();
         for (i = 0; i < jk_shmem.hdr->h.data.pos; i += JK_SHM_SLOT_SIZE) {
@@ -811,6 +843,9 @@ jk_shm_worker_header_t *jk_shm_alloc_worker(jk_pool_t *p, int type,
         else {
             /* No more free memory left.
              */
+            jk_log(l, JK_LOG_ERROR,
+                   "Could not allocate shared memory for worker %s",
+                   name);
             w = NULL;
         }
         jk_shm_unlock();
@@ -834,38 +869,21 @@ const char *jk_shm_name()
     return jk_shmem.filename;
 }
 
-
-time_t jk_shm_get_workers_time()
+int jk_shm_check_maintain(time_t trigger)
 {
-    if (jk_shmem.hdr)
-        return jk_shmem.hdr->h.data.modified;
-    else
-        return jk_workers_modified_time;
-}
-
-void jk_shm_set_workers_time(time_t t)
-{
-    jk_shm_lock();
-    if (jk_shmem.hdr)
-        jk_shmem.hdr->h.data.modified = t;
-    else
-        jk_workers_modified_time = t;
-    jk_workers_access_time = t;
-    jk_shm_unlock();
-}
-
-int jk_shm_is_modified()
-{
-    time_t m = jk_shm_get_workers_time();
-    if (m != jk_workers_access_time)
-        return 1;
-    else
-        return 0;
-}
-
-void jk_shm_sync_access_time()
-{
-    jk_workers_access_time = jk_shm_get_workers_time();
+    int rv = JK_FALSE;
+    int maintain_checking = JK_ATOMIC_INCREMENT(&(jk_shmem.hdr->h.data.maintain_checking));
+    /* Another process (or thread) is already checking */
+    if (maintain_checking > 1) {
+        JK_ATOMIC_DECREMENT(&(jk_shmem.hdr->h.data.maintain_checking));
+        return rv;
+    }
+    if (jk_shmem.hdr->h.data.maintain_time < trigger) {
+        jk_shmem.hdr->h.data.maintain_time = time(NULL);
+        rv = JK_TRUE;
+    }
+    JK_ATOMIC_DECREMENT(&(jk_shmem.hdr->h.data.maintain_checking));
+    return rv;
 }
 
 int jk_shm_lock()
@@ -907,20 +925,23 @@ int jk_shm_unlock()
     return rc;
 }
 
-jk_shm_ajp_worker_t *jk_shm_alloc_ajp_worker(jk_pool_t *p, const char *name)
+jk_shm_ajp_worker_t *jk_shm_alloc_ajp_worker(jk_pool_t *p, const char *name,
+                                             jk_logger_t *l)
 {
     return (jk_shm_ajp_worker_t *)jk_shm_alloc_worker(p,
-                                    JK_AJP13_WORKER_TYPE, 0, name);
+                                    JK_AJP13_WORKER_TYPE, 0, name, l);
 }
 
-jk_shm_lb_sub_worker_t *jk_shm_alloc_lb_sub_worker(jk_pool_t *p, int lb_id, const char *name)
+jk_shm_lb_sub_worker_t *jk_shm_alloc_lb_sub_worker(jk_pool_t *p, int lb_id, const char *name,
+                                                   jk_logger_t *l)
 {
     return (jk_shm_lb_sub_worker_t *)jk_shm_alloc_worker(p,
-                                    JK_LB_SUB_WORKER_TYPE, lb_id, name);
+                                    JK_LB_SUB_WORKER_TYPE, lb_id, name, l);
 }
 
-jk_shm_lb_worker_t *jk_shm_alloc_lb_worker(jk_pool_t *p, const char *name)
+jk_shm_lb_worker_t *jk_shm_alloc_lb_worker(jk_pool_t *p, const char *name,
+                                           jk_logger_t *l)
 {
     return (jk_shm_lb_worker_t *)jk_shm_alloc_worker(p,
-                                    JK_LB_WORKER_TYPE, 0, name);
+                                    JK_LB_WORKER_TYPE, 0, name, l);
 }
