@@ -20,7 +20,7 @@
  * Author:      Gal Shachor <shachor@il.ibm.com>                           *
  * Author:      Henri Gomez <hgomez@apache.org>                            *
  * Author:      Rainer Jung <rjung@apache.org>                             *
- * Version:     $Revision: 1338797 $                                          *
+ * Version:     $Revision: 1658173 $                                          *
  ***************************************************************************/
 
 
@@ -45,6 +45,7 @@
 #define NATIVE_LIB_OF_WORKER        "native_lib"
 #define REFERENCE_OF_WORKER         "reference"
 #define HOST_OF_WORKER              "host"
+#define SOURCE_OF_WORKER            "source"
 #define PORT_OF_WORKER              "port"
 #define TYPE_OF_WORKER              "type"
 #define CACHE_OF_WORKER_DEPRECATED  "cachesize"
@@ -73,6 +74,8 @@
 #define STICKY_SESSION_FORCE        "sticky_session_force"
 #define SESSION_COOKIE_OF_WORKER    "session_cookie"
 #define SESSION_PATH_OF_WORKER      "session_path"
+#define SET_SESSION_COOKIE          "set_session_cookie"
+#define SESSION_COOKIE_PATH_OF_WORKER "session_cookie_path"
 
 #define LOCAL_WORKER_DEPRECATED     "local_worker"
 #define LOCAL_WORKER_ONLY_DEPRECATED "local_worker_only"
@@ -90,6 +93,7 @@
 #define WORKER_ERROR_ESCALATION_TIME "error_escalation_time"
 #define MAX_REPLY_TIMEOUTS_OF_WORKER "max_reply_timeouts"
 #define RETRY_INTERVAL_OF_WORKER    "retry_interval"
+#define BUSY_LIMIT_OF_WORKER        "busy_limit"
 #define WORKER_MAX_PACKET_SIZE      "max_packet_size"
 #define STYLE_SHEET_OF_WORKER       "css"
 #define NAMESPACE_OF_WORKER         "ns"
@@ -102,6 +106,7 @@
 #define USER_CASE_OF_WORKER         "user_case_insensitive"
 #define GOOD_RATING_OF_WORKER       "good"
 #define BAD_RATING_OF_WORKER        "bad"
+#define PREFER_IPV6_ADDRESS         "prefer_ipv6"
 
 #define DEFAULT_WORKER_TYPE         JK_AJP13_WORKER_NAME
 #define SECRET_KEY_OF_WORKER        "secretkey"
@@ -123,13 +128,23 @@
 #define TOMCAT41_BRIDGE_NAME        "tomcat41"
 #define TOMCAT50_BRIDGE_NAME        "tomcat5"
 
-#define HUGE_BUFFER_SIZE 8192
+#define LOG_BUFFER_SIZE 1024
 
-#define MAKE_WORKER_PARAM(P)     \
-        strcpy(buf, "worker.");  \
-        strcat(buf, wname);      \
-        strcat(buf, ".");        \
-        strcat(buf, P)
+/*
+ * Our longest worker attribute name is below 30 bytes.
+ * Add space for "worker.", another ".", the
+ * worker name and the final '\0'.
+ */
+#define JK_MAX_ATTRIBUTE_NAME_LEN (30)
+#define PARAM_BUFFER_SIZE (JK_MAX_NAME_LEN + 8 + JK_MAX_ATTRIBUTE_NAME_LEN + 1)
+#define MAKE_WORKER_PARAM(P) \
+        { \
+            size_t remain = PARAM_BUFFER_SIZE; \
+            strcpy(buf, "worker."); remain -= strlen("worker."); \
+            strncat(buf, wname, remain); remain -= strlen(wname); \
+            strncat(buf, ".", remain); remain -= 1; \
+            strncat(buf, P, remain); \
+        }
 
 /*
  * define the log format, we're using by default the one from error.log
@@ -161,7 +176,7 @@ static const char *list_properties[] = {
     GOOD_RATING_OF_WORKER,
     BAD_RATING_OF_WORKER,
     STATUS_FAIL_OF_WORKER,
-    "list",
+    LIST_PROPERTY_NAME,
     NULL
 };
 
@@ -169,6 +184,7 @@ static const char *unique_properties[] = {
     SECRET_OF_WORKER,
     REFERENCE_OF_WORKER,
     HOST_OF_WORKER,
+    SOURCE_OF_WORKER,
     PORT_OF_WORKER,
     TYPE_OF_WORKER,
     CACHE_OF_WORKER_DEPRECATED,
@@ -194,6 +210,8 @@ static const char *unique_properties[] = {
     STICKY_SESSION_FORCE,
     SESSION_COOKIE_OF_WORKER,
     SESSION_PATH_OF_WORKER,
+    SET_SESSION_COOKIE,
+    SESSION_COOKIE_PATH_OF_WORKER,
     LOCAL_WORKER_DEPRECATED,
     LOCAL_WORKER_ONLY_DEPRECATED,
     JVM_ROUTE_OF_WORKER_DEPRECATED,
@@ -209,6 +227,7 @@ static const char *unique_properties[] = {
     WORKER_ERROR_ESCALATION_TIME,
     MAX_REPLY_TIMEOUTS_OF_WORKER,
     RETRY_INTERVAL_OF_WORKER,
+    BUSY_LIMIT_OF_WORKER,
     WORKER_MAX_PACKET_SIZE,
     STYLE_SHEET_OF_WORKER,
     READ_ONLY_OF_WORKER,
@@ -219,6 +238,7 @@ static const char *unique_properties[] = {
     XML_DOCTYPE_OF_WORKER,
     PROP_PREFIX_OF_WORKER,
     USER_CASE_OF_WORKER,
+    PREFER_IPV6_ADDRESS,
     NULL
 };
 
@@ -261,6 +281,7 @@ static const char *supported_properties[] = {
     NATIVE_LIB_OF_WORKER,
     REFERENCE_OF_WORKER,
     HOST_OF_WORKER,
+    SOURCE_OF_WORKER,
     PORT_OF_WORKER,
     TYPE_OF_WORKER,
     CACHE_OF_WORKER_DEPRECATED,
@@ -289,6 +310,8 @@ static const char *supported_properties[] = {
     STICKY_SESSION_FORCE,
     SESSION_COOKIE_OF_WORKER,
     SESSION_PATH_OF_WORKER,
+    SET_SESSION_COOKIE,
+    SESSION_COOKIE_PATH_OF_WORKER,
     LOCAL_WORKER_DEPRECATED,
     LOCAL_WORKER_ONLY_DEPRECATED,
     JVM_ROUTE_OF_WORKER_DEPRECATED,
@@ -305,6 +328,7 @@ static const char *supported_properties[] = {
     WORKER_ERROR_ESCALATION_TIME,
     MAX_REPLY_TIMEOUTS_OF_WORKER,
     RETRY_INTERVAL_OF_WORKER,
+    BUSY_LIMIT_OF_WORKER,
     WORKER_MAX_PACKET_SIZE,
     STYLE_SHEET_OF_WORKER,
     NAMESPACE_OF_WORKER,
@@ -321,6 +345,7 @@ static const char *supported_properties[] = {
     STATUS_FAIL_OF_WORKER,
     LIST_PROPERTY_NAME,
     MAINTAIN_PROPERTY_NAME,
+    PREFER_IPV6_ADDRESS,
     NULL
 };
 
@@ -351,13 +376,13 @@ int jk_get_bool_code(const char *v, int def)
              *v == 'F' || *v == 'f' ||
              *v == 'N' || *v == 'n' ||
             (*v == '0' && *(v + 1) == '\0')) {
-        return 0;
+        return JK_FALSE;
     }
     if (!strcasecmp(v, "on") ||
              *v == 'T' || *v == 't' ||
              *v == 'Y' || *v == 'y' ||
             (*v == '1' && *(v + 1) == '\0')) {
-        return 1;
+        return JK_TRUE;
     }
     return def;
 }
@@ -674,12 +699,13 @@ int jk_log(jk_logger_t *l,
            const char *fmt, ...)
 {
     int rc = 0;
+
     /*
      * Need to reserve space for terminating zero byte
      * and platform specific line endings added during the call
      * to the output routing.
      */
-    static int usable_size = HUGE_BUFFER_SIZE - 3;
+    static int usable_size = LOG_BUFFER_SIZE - 3;
     if (!l || !file || !fmt) {
         return -1;
     }
@@ -690,7 +716,7 @@ int jk_log(jk_logger_t *l,
         /* we will allocate and free the temporary buffer in this function         */
         char *buf;
 #else
-        char buf[HUGE_BUFFER_SIZE];
+        char buf[LOG_BUFFER_SIZE];
 #endif
         char *f = (char *)(file + strlen(file) - 1);
         va_list args;
@@ -704,7 +730,7 @@ int jk_log(jk_logger_t *l,
         }
 
 #ifdef NETWARE
-        buf = (char *)malloc(HUGE_BUFFER_SIZE);
+        buf = (char *)malloc(LOG_BUFFER_SIZE);
         if (NULL == buf)
             return -1;
 #endif
@@ -720,6 +746,8 @@ int jk_log(jk_logger_t *l,
                           "[%" JK_PID_T_FMT ":%" JK_PTHREAD_T_FMT "] ", getpid(), jk_gettid());
             used += rc;
             if (rc < 0 ) {
+                strcpy(buf, "Logging failed in pid/tid formatting");
+                l->log(l, level, strlen(buf), buf);
                 return 0;
             }
 
@@ -729,6 +757,8 @@ int jk_log(jk_logger_t *l,
                 used += rc;
             }
             else {
+                strcpy(buf, "Logging failed in log level formatting");
+                l->log(l, level, strlen(buf), buf);
                 return 0;           /* [V] not sure what to return... */
             }
 
@@ -741,6 +771,8 @@ int jk_log(jk_logger_t *l,
                     used += 2;
                 }
                 else {
+                    strcpy(buf, "Logging failed in function name formatting");
+                    l->log(l, level, strlen(buf), buf);
                     return 0;           /* [V] not sure what to return... */
                 }
             }
@@ -751,6 +783,8 @@ int jk_log(jk_logger_t *l,
                 used += rc;
             }
             else {
+                strcpy(buf, "Logging failed in source file name formatting");
+                l->log(l, level, strlen(buf), buf);
                 return 0;           /* [V] not sure what to return... */
             }
 
@@ -758,6 +792,8 @@ int jk_log(jk_logger_t *l,
                           " (%d): ", line);
             used += rc;
             if (rc < 0 || usable_size - used < 0) {
+                strcpy(buf, "Logging failed in line number formatting");
+                l->log(l, level, strlen(buf), buf);
                 return 0;           /* [V] not sure what to return... */
             }
         }
@@ -765,11 +801,16 @@ int jk_log(jk_logger_t *l,
         va_start(args, fmt);
         rc = vsnprintf(buf + used, usable_size - used, fmt, args);
         va_end(args);
-        if ( rc <= usable_size - used ) {
+        /* Depending on the snprintf implementation used,
+         * "rc == usable_size - used" can indicate not enough space in buffer */
+        if (rc < usable_size - used) {
             used += rc;
         }
         else {
             used = usable_size;
+            buf[used - 1] = '.';
+            buf[used - 2] = '.';
+            buf[used - 3] = '.';
         }
         l->log(l, level, used, buf);
 
@@ -781,9 +822,23 @@ int jk_log(jk_logger_t *l,
     return rc;
 }
 
+int jk_check_attribute_length(const char *name, const char *value,
+                              jk_logger_t *l)
+{
+    size_t len = strlen(value);
+    if (len > JK_MAX_NAME_LEN) {
+        jk_log(l, JK_LOG_ERROR,
+               "Worker %s '%s' is %d bytes too long, "
+               "a maximum of %d bytes is supported",
+               name, value, len - JK_MAX_NAME_LEN, JK_MAX_NAME_LEN);
+        return JK_FALSE;
+    }
+    return JK_TRUE;
+}
+
 const char *jk_get_worker_type(jk_map_t *m, const char *wname)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
         return NULL;
@@ -794,13 +849,13 @@ const char *jk_get_worker_type(jk_map_t *m, const char *wname)
 
 const char *jk_get_worker_route(jk_map_t *m, const char *wname, const char *def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     const char *v;
     if (!m || !wname) {
-        return NULL;
+        return def;
     }
     MAKE_WORKER_PARAM(ROUTE_OF_WORKER);
-    v = jk_map_get_string(m, buf, NULL);
+    v = jk_map_get_string(m, buf, def);
     if (v) {
         return v;
     }
@@ -811,9 +866,9 @@ const char *jk_get_worker_route(jk_map_t *m, const char *wname, const char *def)
 
 const char *jk_get_worker_domain(jk_map_t *m, const char *wname, const char *def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     if (!m || !wname) {
-        return NULL;
+        return def;
     }
     MAKE_WORKER_PARAM(DOMAIN_OF_WORKER);
     return jk_map_get_string(m, buf, def);
@@ -821,9 +876,9 @@ const char *jk_get_worker_domain(jk_map_t *m, const char *wname, const char *def
 
 const char *jk_get_worker_redirect(jk_map_t *m, const char *wname, const char *def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     if (!m || !wname) {
-        return NULL;
+        return def;
     }
     MAKE_WORKER_PARAM(REDIRECT_OF_WORKER);
     return jk_map_get_string(m, buf, def);
@@ -831,7 +886,7 @@ const char *jk_get_worker_redirect(jk_map_t *m, const char *wname, const char *d
 
 const char *jk_get_worker_secret(jk_map_t *m, const char *wname)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
         return NULL;
@@ -842,45 +897,9 @@ const char *jk_get_worker_secret(jk_map_t *m, const char *wname)
     return jk_map_get_string(m, buf, NULL);
 }
 
-/* [V] I suggest that the following general purpose functions be used.       */
-/*     More should be added (double etc.), but now these were enough for me. */
-/*     Functions that can be simulated with these should be "deprecated".    */
-
-int jk_get_worker_str_prop(jk_map_t *m,
-                           const char *wname, const char *pname, const char **prop)
-{
-    char buf[1024];
-
-    if (m && prop && wname && pname) {
-        MAKE_WORKER_PARAM(pname);
-        *prop = jk_map_get_string(m, buf, NULL);
-        if (*prop) {
-            return JK_TRUE;
-        }
-    }
-    return JK_FALSE;
-}
-
-int jk_get_worker_int_prop(jk_map_t *m,
-                           const char *wname, const char *pname, int *prop)
-{
-    char buf[1024];
-
-    if (m && prop && wname && pname) {
-        int i;
-        MAKE_WORKER_PARAM(pname);
-        i = jk_map_get_int(m, buf, -1);
-        if (-1 != i) {
-            *prop = i;
-            return JK_TRUE;
-        }
-    }
-    return JK_FALSE;
-}
-
 const char *jk_get_worker_host(jk_map_t *m, const char *wname, const char *def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
         return NULL;
@@ -891,12 +910,25 @@ const char *jk_get_worker_host(jk_map_t *m, const char *wname, const char *def)
     return jk_map_get_string(m, buf, def);
 }
 
-int jk_get_worker_port(jk_map_t *m, const char *wname, int def)
+const char *jk_get_worker_source(jk_map_t *m, const char *wname, const char *def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return NULL;
+    }
+
+    MAKE_WORKER_PARAM(SOURCE_OF_WORKER);
+
+    return jk_map_get_string(m, buf, def);
+}
+
+int jk_get_worker_port(jk_map_t *m, const char *wname, int def)
+{
+    char buf[PARAM_BUFFER_SIZE];
+
+    if (!m || !wname) {
+        return def;
     }
 
     MAKE_WORKER_PARAM(PORT_OF_WORKER);
@@ -904,6 +936,18 @@ int jk_get_worker_port(jk_map_t *m, const char *wname, int def)
     return jk_map_get_int(m, buf, def);
 }
 
+int jk_get_worker_prefer_ipv6(jk_map_t *m, const char *wname, int def)
+{
+    char buf[PARAM_BUFFER_SIZE];
+    if (!m || !wname) {
+        return def;
+    }
+
+    MAKE_WORKER_PARAM(PREFER_IPV6_ADDRESS);
+
+    return jk_map_get_bool(m, buf, def);
+}
+
 static int def_cache_size = -1;
 int jk_get_worker_def_cache_size(int protocol)
 {
@@ -923,11 +967,11 @@ void jk_set_worker_def_cache_size(int sz)
 
 int jk_get_worker_cache_size(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     int rv;
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(CACHE_OF_WORKER);
@@ -939,10 +983,10 @@ int jk_get_worker_cache_size(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_cache_size_min(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(CACHE_OF_WORKER_MIN);
@@ -951,10 +995,10 @@ int jk_get_worker_cache_size_min(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_cache_acquire_timeout(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(CACHE_ACQUIRE_OF_WORKER);
@@ -963,10 +1007,10 @@ int jk_get_worker_cache_acquire_timeout(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_socket_timeout(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(SOCKET_TIMEOUT_OF_WORKER);
@@ -976,10 +1020,10 @@ int jk_get_worker_socket_timeout(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_socket_connect_timeout(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(SOCKET_CONNECT_TIMEOUT_OF_WORKER);
@@ -989,10 +1033,10 @@ int jk_get_worker_socket_connect_timeout(jk_map_t *m, const char *wname, int def
 
 int jk_get_worker_recover_timeout(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(WORKER_RECOVER_TIME);
@@ -1002,10 +1046,10 @@ int jk_get_worker_recover_timeout(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_error_escalation_time(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(WORKER_ERROR_ESCALATION_TIME);
@@ -1015,10 +1059,10 @@ int jk_get_worker_error_escalation_time(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_max_reply_timeouts(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(MAX_REPLY_TIMEOUTS_OF_WORKER);
@@ -1028,10 +1072,10 @@ int jk_get_worker_max_reply_timeouts(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_retry_interval(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(RETRY_INTERVAL_OF_WORKER);
@@ -1039,12 +1083,25 @@ int jk_get_worker_retry_interval(jk_map_t *m, const char *wname, int def)
     return jk_map_get_int(m, buf, def);
 }
 
+int jk_get_worker_busy_limit(jk_map_t *m, const char *wname, int def)
+{
+    char buf[PARAM_BUFFER_SIZE];
+
+    if (!m || !wname) {
+        return def;
+    }
+
+    MAKE_WORKER_PARAM(BUSY_LIMIT_OF_WORKER);
+
+    return jk_map_get_int(m, buf, def);
+}
+
 int jk_get_worker_socket_buffer(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     int i;
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(SOCKET_BUFFER_OF_WORKER);
@@ -1057,10 +1114,9 @@ int jk_get_worker_socket_buffer(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_socket_keepalive(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
-
+    char buf[PARAM_BUFFER_SIZE];
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(SOCKET_KEEPALIVE_OF_WORKER);
@@ -1070,10 +1126,10 @@ int jk_get_worker_socket_keepalive(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_conn_ping_interval(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(CONN_PING_INTERVAL_OF_WORKER);
@@ -1083,11 +1139,11 @@ int jk_get_worker_conn_ping_interval(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_cache_timeout(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     int rv;
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(CACHE_TIMEOUT_OF_WORKER);
@@ -1100,10 +1156,10 @@ int jk_get_worker_cache_timeout(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_connect_timeout(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(CONNECT_TIMEOUT_OF_WORKER);
@@ -1113,10 +1169,10 @@ int jk_get_worker_connect_timeout(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_prepost_timeout(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(PREPOST_TIMEOUT_OF_WORKER);
@@ -1126,10 +1182,10 @@ int jk_get_worker_prepost_timeout(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_ping_timeout(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(PING_TIMEOUT_OF_WORKER);
@@ -1139,7 +1195,8 @@ int jk_get_worker_ping_timeout(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_ping_mode(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
+    char mode[100];
     const char *v;
 
     if (!m || !wname) {
@@ -1148,16 +1205,17 @@ int jk_get_worker_ping_mode(jk_map_t *m, const char *wname, int def)
 
     MAKE_WORKER_PARAM(PING_MODE_OF_WORKER);
 
-    v = jk_map_get_string(m, buf, NULL);
+    jk_ajp_get_cping_text(def, mode);
+    v = jk_map_get_string(m, buf, mode);
     return jk_ajp_get_cping_mode(v, def);
 }
 
 int jk_get_worker_reply_timeout(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(REPLY_TIMEOUT_OF_WORKER);
@@ -1172,10 +1230,10 @@ int jk_get_worker_recycle_timeout(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_retries(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     int rv;
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(RETRIES_OF_WORKER);
@@ -1189,10 +1247,10 @@ int jk_get_worker_retries(jk_map_t *m, const char *wname, int def)
 
 int jk_get_worker_recovery_opts(jk_map_t *m, const char *wname, int def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return -1;
+        return def;
     }
 
     MAKE_WORKER_PARAM(RECOVERY_OPTS_OF_WORKER);
@@ -1202,7 +1260,7 @@ int jk_get_worker_recovery_opts(jk_map_t *m, const char *wname, int def)
 
 const char *jk_get_worker_secret_key(jk_map_t *m, const char *wname)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
         return NULL;
@@ -1232,37 +1290,31 @@ int jk_get_worker_list(jk_map_t *m, char ***list, unsigned *num_of_workers)
 
 int jk_get_is_worker_disabled(jk_map_t *m, const char *wname)
 {
-    int rc = JK_TRUE;
-    char buf[1024];
+    int def = JK_FALSE;
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && wname) {
-        int value;
         MAKE_WORKER_PARAM(IS_WORKER_DISABLED_DEPRECATED);
-        value = jk_map_get_bool(m, buf, 0);
-        if (!value)
-            rc = JK_FALSE;
+        return jk_map_get_bool(m, buf, def);
     }
-    return rc;
+    return JK_TRUE;
 }
 
 int jk_get_is_worker_stopped(jk_map_t *m, const char *wname)
 {
-    int rc = JK_TRUE;
-    char buf[1024];
+    int def = JK_FALSE;
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && wname) {
-        int value;
         MAKE_WORKER_PARAM(IS_WORKER_STOPPED_DEPRECATED);
-        value = jk_map_get_bool(m, buf, 0);
-        if (!value)
-            rc = JK_FALSE;
+        return jk_map_get_bool(m, buf, def);
     }
-    return rc;
+    return JK_TRUE;
 }
 
 int jk_get_worker_activation(jk_map_t *m, const char *wname)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     const char *v;
 
     if (!m || !wname) {
@@ -1282,7 +1334,7 @@ int jk_get_worker_activation(jk_map_t *m, const char *wname)
 
 int jk_get_lb_factor(jk_map_t *m, const char *wname)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
         return DEFAULT_LB_FACTOR;
@@ -1295,7 +1347,7 @@ int jk_get_lb_factor(jk_map_t *m, const char *wname)
 
 int jk_get_distance(jk_map_t *m, const char *wname)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
         return DEFAULT_DISTANCE;
@@ -1308,37 +1360,31 @@ int jk_get_distance(jk_map_t *m, const char *wname)
 
 int jk_get_is_sticky_session(jk_map_t *m, const char *wname)
 {
-    int rc = JK_TRUE;
-    char buf[1024];
+    int def = JK_TRUE;
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && wname) {
-        int value;
         MAKE_WORKER_PARAM(STICKY_SESSION);
-        value = jk_map_get_bool(m, buf, 1);
-        if (!value)
-            rc = JK_FALSE;
+        return jk_map_get_bool(m, buf, def);
     }
-    return rc;
+    return def;
 }
 
 int jk_get_is_sticky_session_force(jk_map_t *m, const char *wname)
 {
-    int rc = JK_FALSE;
-    char buf[1024];
+    int def = JK_FALSE;
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && wname) {
-        int value;
         MAKE_WORKER_PARAM(STICKY_SESSION_FORCE);
-        value = jk_map_get_bool(m, buf, 0);
-        if (value)
-            rc = JK_TRUE;
+        return jk_map_get_bool(m, buf, def);
     }
-    return rc;
+    return def;
 }
 
 int jk_get_lb_method(jk_map_t *m, const char *wname)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     const char *v;
 
     if (!m || !wname) {
@@ -1346,13 +1392,13 @@ int jk_get_lb_method(jk_map_t *m, const char *wname)
     }
 
     MAKE_WORKER_PARAM(METHOD_OF_WORKER);
-    v = jk_map_get_string(m, buf, NULL);
+    v = jk_map_get_string(m, buf, JK_LB_METHOD_DEF);
     return jk_lb_get_method_code(v);
 }
 
 int jk_get_lb_lock(jk_map_t *m, const char *wname)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     const char *v;
 
     if (!m || !wname) {
@@ -1360,70 +1406,72 @@ int jk_get_lb_lock(jk_map_t *m, const char *wname)
     }
 
     MAKE_WORKER_PARAM(LOCK_OF_WORKER);
-    v = jk_map_get_string(m, buf, NULL);
+    v = jk_map_get_string(m, buf, JK_LB_LOCK_DEF);
     return jk_lb_get_lock_code(v);
 }
 
 int jk_get_max_packet_size(jk_map_t *m, const char *wname)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     int sz;
 
     if (!m || !wname) {
-        return DEF_BUFFER_SZ;
+        return AJP13_DEF_PACKET_SIZE;
     }
 
     MAKE_WORKER_PARAM(WORKER_MAX_PACKET_SIZE);
-    sz = jk_map_get_int(m, buf, DEF_BUFFER_SZ);
-    sz = JK_ALIGN(sz, 1024);
-    if (sz < DEF_BUFFER_SZ)
-        sz = DEF_BUFFER_SZ;
-    else if (sz > 64*1024)
-        sz = 64*1024;
+    sz = jk_map_get_int(m, buf, AJP13_DEF_PACKET_SIZE);
+    sz = JK_ALIGN(sz, AJP13_PACKET_SIZE_ALIGN);
+    if (sz < AJP13_DEF_PACKET_SIZE)
+        sz = AJP13_DEF_PACKET_SIZE;
+    else if (sz > AJP13_MAX_PACKET_SIZE)
+        sz = AJP13_MAX_PACKET_SIZE;
 
     return sz;
 }
 
 int jk_get_worker_fail_on_status(jk_map_t *m, const char *wname,
-                                 int *list, unsigned int list_size)
+                                 int **list, unsigned int *list_size)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
+    int *ar;
 
-    if (!m || !wname || !list) {
+    if (!m || !wname || !list || !list_size) {
         return 0;
     }
     MAKE_WORKER_PARAM(STATUS_FAIL_OF_WORKER);
-    if (list_size) {
-        return jk_map_get_int_list(m, buf,
-                                   list, list_size,
-                                   NULL);
+    ar = jk_map_get_int_list(m,
+                             buf,
+                             list_size,
+                             NULL);
+    if (ar) {
+        *list = ar;
+        return JK_TRUE;
     }
+    *list = NULL;
+    *list_size = 0;
 
-    return 0;
+    return JK_FALSE;
 }
 
 int jk_get_worker_user_case_insensitive(jk_map_t *m, const char *wname)
 {
-    int rc = JK_FALSE;
-    char buf[1024];
+    int def = JK_FALSE;
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && wname) {
-        int value;
         MAKE_WORKER_PARAM(USER_CASE_OF_WORKER);
-        value = jk_map_get_bool(m, buf, 0);
-        if (value)
-            rc = JK_TRUE;
+        return jk_map_get_bool(m, buf, def);
     }
-    return rc;
-
+    return def;
 }
 
 const char *jk_get_worker_style_sheet(jk_map_t *m, const char *wname, const char *def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return NULL;
+        return def;
     }
 
     MAKE_WORKER_PARAM(STYLE_SHEET_OF_WORKER);
@@ -1434,10 +1482,10 @@ const char *jk_get_worker_style_sheet(jk_map_t *m, const char *wname, const char
 const char *jk_get_worker_name_space(jk_map_t *m, const char *wname, const char *def)
 {
     const char *rc;
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return NULL;
+        return def;
     }
     MAKE_WORKER_PARAM(NAMESPACE_OF_WORKER);
     rc = jk_map_get_string(m, buf, def);
@@ -1450,10 +1498,10 @@ const char *jk_get_worker_name_space(jk_map_t *m, const char *wname, const char
 const char *jk_get_worker_xmlns(jk_map_t *m, const char *wname, const char *def)
 {
     const char *rc;
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return NULL;
+        return def;
     }
     MAKE_WORKER_PARAM(XML_NAMESPACE_OF_WORKER);
     rc = jk_map_get_string(m, buf, def);
@@ -1465,10 +1513,10 @@ const char *jk_get_worker_xmlns(jk_map_t *m, const char *wname, const char *def)
 
 const char *jk_get_worker_xml_doctype(jk_map_t *m, const char *wname, const char *def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return NULL;
+        return def;
     }
     MAKE_WORKER_PARAM(XML_DOCTYPE_OF_WORKER);
     return jk_map_get_string(m, buf, def);
@@ -1476,10 +1524,10 @@ const char *jk_get_worker_xml_doctype(jk_map_t *m, const char *wname, const char
 
 const char *jk_get_worker_prop_prefix(jk_map_t *m, const char *wname, const char *def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
-        return NULL;
+        return def;
     }
     MAKE_WORKER_PARAM(PROP_PREFIX_OF_WORKER);
     return jk_map_get_string(m, buf, def);
@@ -1487,24 +1535,21 @@ const char *jk_get_worker_prop_prefix(jk_map_t *m, const char *wname, const char
 
 int jk_get_is_read_only(jk_map_t *m, const char *wname)
 {
-    int rc = JK_FALSE;
-    char buf[1024];
+    int def = JK_FALSE;
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && wname) {
-        int value;
         MAKE_WORKER_PARAM(READ_ONLY_OF_WORKER);
-        value = jk_map_get_bool(m, buf, 0);
-        if (value)
-            rc = JK_TRUE;
+        return jk_map_get_bool(m, buf, def);
     }
-    return rc;
+    return def;
 }
 
 int jk_get_worker_user_list(jk_map_t *m,
                             const char *wname,
                             char ***list, unsigned int *num)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && list && num && wname) {
         char **ar = NULL;
@@ -1526,7 +1571,7 @@ int jk_get_worker_good_rating(jk_map_t *m,
                               const char *wname,
                               char ***list, unsigned int *num)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && list && num && wname) {
         char **ar = NULL;
@@ -1548,7 +1593,7 @@ int jk_get_worker_bad_rating(jk_map_t *m,
                              const char *wname,
                              char ***list, unsigned int *num)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && list && num && wname) {
         char **ar = NULL;
@@ -1570,7 +1615,7 @@ int jk_get_lb_worker_list(jk_map_t *m,
                           const char *wname,
                           char ***list, unsigned int *num_of_workers)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && list && num_of_workers && wname) {
         char **ar = NULL;
@@ -1599,7 +1644,7 @@ int jk_get_worker_mount_list(jk_map_t *m,
                              const char *wname,
                              char ***list, unsigned int *num_of_maps)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && list && num_of_maps && wname) {
         char **ar = NULL;
@@ -1625,7 +1670,7 @@ int jk_get_worker_maintain_time(jk_map_t *m)
 
 int jk_get_worker_mx(jk_map_t *m, const char *wname, unsigned *mx)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && mx && wname) {
         int i;
@@ -1643,7 +1688,7 @@ int jk_get_worker_mx(jk_map_t *m, const char *wname, unsigned *mx)
 
 int jk_get_worker_ms(jk_map_t *m, const char *wname, unsigned *ms)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && ms && wname) {
         int i;
@@ -1661,7 +1706,7 @@ int jk_get_worker_ms(jk_map_t *m, const char *wname, unsigned *ms)
 
 int jk_get_worker_classpath(jk_map_t *m, const char *wname, const char **cp)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && cp && wname) {
         MAKE_WORKER_PARAM(CP_OF_WORKER);
@@ -1677,7 +1722,7 @@ int jk_get_worker_classpath(jk_map_t *m, const char *wname, const char **cp)
 
 int jk_get_worker_bridge_type(jk_map_t *m, const char *wname, unsigned *bt)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
     const char *type;
 
     if (m && bt && wname) {
@@ -1706,7 +1751,7 @@ int jk_get_worker_bridge_type(jk_map_t *m, const char *wname, unsigned *bt)
 
 int jk_get_worker_jvm_path(jk_map_t *m, const char *wname, const char **vm_path)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && vm_path && wname) {
         MAKE_WORKER_PARAM(JVM_OF_WORKER);
@@ -1723,7 +1768,7 @@ int jk_get_worker_jvm_path(jk_map_t *m, const char *wname, const char **vm_path)
 /* [V] This is unused. currently. */
 int jk_get_worker_callback_dll(jk_map_t *m, const char *wname, const char **cb_path)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && cb_path && wname) {
         MAKE_WORKER_PARAM(NATIVE_LIB_OF_WORKER);
@@ -1739,7 +1784,7 @@ int jk_get_worker_callback_dll(jk_map_t *m, const char *wname, const char **cb_p
 
 int jk_get_worker_cmd_line(jk_map_t *m, const char *wname, const char **cmd_line)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && cmd_line && wname) {
         MAKE_WORKER_PARAM(CMD_LINE_OF_WORKER);
@@ -1789,18 +1834,15 @@ int jk_file_exists(const char *f)
 
 static int jk_is_some_property(const char *prp_name, const char *suffix, const char *sep)
 {
-    char buf[1024];
-
-    if (prp_name && suffix) {
-        size_t prp_name_len;
-        size_t suffix_len;
-        strcpy(buf, sep);
-        strcat(buf, suffix);
-        prp_name_len = strlen(prp_name);
-        suffix_len = strlen(buf);
-        if (prp_name_len >= suffix_len) {
-            const char *prp_suffix = prp_name + prp_name_len - suffix_len;
-            if (!strcmp(buf, prp_suffix)) {
+    if (prp_name && suffix && sep) {
+        size_t prp_name_len = strlen(prp_name);
+        size_t suffix_len = strlen(suffix);
+        size_t sep_len = strlen(sep);
+        size_t sep_off = sep_len + suffix_len;
+        
+        if (prp_name_len >= sep_off) {
+            if (!strncmp(prp_name + prp_name_len - sep_off, sep, sep_len) &&
+                !strncmp(prp_name + prp_name_len - suffix_len, suffix, suffix_len)) {
                 return JK_TRUE;
             }
         }
@@ -1854,6 +1896,23 @@ int jk_is_deprecated_property(const char *prp_name)
     }
     return JK_FALSE;
 }
+
+int jk_check_buffer_size()
+{
+    const char **props;
+    size_t len = 0;
+    size_t max_len = 0;
+
+    props = &supported_properties[0];
+    while (*props) {
+        len = strlen(*props);
+        if (len > max_len)
+            max_len = len;
+        props++;
+    }
+    return JK_MAX_ATTRIBUTE_NAME_LEN - max_len;
+}
+
 /*
  * Check that property is a valid one (to prevent user typos).
  * Only property starting with worker.
@@ -1862,6 +1921,9 @@ int jk_is_valid_property(const char *prp_name)
 {
     const char **props;
 
+    /* Any property not starting with "worker." is "valid".
+     * It is interpreted as the definition of a custom variable.
+     */
     if (memcmp(prp_name, "worker.", 7))
         return JK_TRUE;
 
@@ -1876,7 +1938,7 @@ int jk_is_valid_property(const char *prp_name)
 
 int jk_get_worker_stdout(jk_map_t *m, const char *wname, const char **stdout_name)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && stdout_name && wname) {
         MAKE_WORKER_PARAM(STDOUT_OF_WORKER);
@@ -1892,7 +1954,7 @@ int jk_get_worker_stdout(jk_map_t *m, const char *wname, const char **stdout_nam
 
 int jk_get_worker_stderr(jk_map_t *m, const char *wname, const char **stderr_name)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && stderr_name && wname) {
         MAKE_WORKER_PARAM(STDERR_OF_WORKER);
@@ -1908,7 +1970,7 @@ int jk_get_worker_stderr(jk_map_t *m, const char *wname, const char **stderr_nam
 
 int jk_get_worker_sysprops(jk_map_t *m, const char *wname, const char **sysprops)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && sysprops && wname) {
         MAKE_WORKER_PARAM(SYSPROPS_OF_WORKER);
@@ -1924,7 +1986,7 @@ int jk_get_worker_sysprops(jk_map_t *m, const char *wname, const char **sysprops
 
 int jk_get_worker_libpath(jk_map_t *m, const char *wname, const char **libpath)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (m && libpath && wname) {
         MAKE_WORKER_PARAM(LIBPATH_OF_WORKER);
@@ -1940,7 +2002,7 @@ int jk_get_worker_libpath(jk_map_t *m, const char *wname, const char **libpath)
 
 const char *jk_get_lb_session_cookie(jk_map_t *m, const char *wname, const char *def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
         return NULL;
@@ -1951,7 +2013,7 @@ const char *jk_get_lb_session_cookie(jk_map_t *m, const char *wname, const char
 
 const char *jk_get_lb_session_path(jk_map_t *m, const char *wname, const char *def)
 {
-    char buf[1024];
+    char buf[PARAM_BUFFER_SIZE];
 
     if (!m || !wname) {
         return NULL;
@@ -1960,6 +2022,27 @@ const char *jk_get_lb_session_path(jk_map_t *m, const char *wname, const char *d
     return jk_map_get_string(m, buf, def);
 }
 
+int jk_get_lb_set_session_cookie(jk_map_t *m, const char *wname, int def)
+{
+    char buf[PARAM_BUFFER_SIZE];
+
+    if (m && wname) {
+        MAKE_WORKER_PARAM(SET_SESSION_COOKIE);
+        return jk_map_get_bool(m, buf, def);
+    }
+    return def;
+}
+
+const char *jk_get_lb_session_cookie_path(jk_map_t *m, const char *wname, const char *def)
+{
+    char buf[PARAM_BUFFER_SIZE];
+
+    if (!m || !wname) {
+        return NULL;
+    }
+    MAKE_WORKER_PARAM(SESSION_COOKIE_PATH_OF_WORKER);
+    return jk_map_get_string(m, buf, def);
+}
 
 int is_http_status_fail(unsigned int http_status_fail_num,
                         int *http_status_fail, int status)
@@ -2089,6 +2172,25 @@ int jk_wildchar_match(const char *str, const char *exp, int icase)
     return (str[x] != '\0');
 }
 
+void jk_no2slash(char *name)
+{
+    char *d, *s;
+
+    s = d = name;
+
+    while (*s) {
+        if ((*d++ = *s) == '/') {
+            do {
+                ++s;
+            } while (*s == '/');
+        }
+        else {
+            ++s;
+        }
+    }
+    *d = '\0';
+}
+
 #ifdef _MT_CODE_PTHREAD
 jk_pthread_t jk_gettid()
 {
