@@ -20,7 +20,7 @@
  * Author:      Gal Shachor <shachor@il.ibm.com>                           *
  * Author:      Henri Gomez <hgomez@apache.org>                            *
  * Author:      Rainer Jung <rjung@apache.org>                             *
- * Version:     $Revision: 1305771 $                                          *
+ * Version:     $Revision: 1658173 $                                          *
  ***************************************************************************/
 #ifndef _JK_UTIL_H
 #define _JK_UTIL_H
@@ -58,19 +58,19 @@ int jk_log(jk_logger_t *l,
            const char *file, int line, const char *funcname, int level,
            const char *fmt, ...);
 
-/* [V] Two general purpose functions. Should ease the function bloat. */
-int jk_get_worker_str_prop(jk_map_t *m,
-                           const char *wname, const char *pname, const char **prop);
-
-int jk_get_worker_int_prop(jk_map_t *m,
-                           const char *wname, const char *pname, int *prop);
-
+int jk_check_attribute_length(const char *name, const char *value,
+                              jk_logger_t *l);
+    
 const char *jk_get_worker_host(jk_map_t *m, const char *wname, const char *def);
 
+const char *jk_get_worker_source(jk_map_t *m, const char *wname, const char *def);
+
 const char *jk_get_worker_type(jk_map_t *m, const char *wname);
 
 int jk_get_worker_port(jk_map_t *m, const char *wname, int def);
 
+int jk_get_worker_prefer_ipv6(jk_map_t *m, const char *wname, int def);
+
 int jk_get_worker_cache_size(jk_map_t *m, const char *wname, int def);
 
 int jk_get_worker_cache_size_min(jk_map_t *m, const char *wname, int def);
@@ -111,6 +111,8 @@ int jk_get_worker_max_reply_timeouts(jk_map_t *m, const char *wname, int def);
 
 int jk_get_worker_retry_interval(jk_map_t *m, const char *wname, int def);
 
+int jk_get_worker_busy_limit(jk_map_t *m, const char *wname, int def);
+
 const char *jk_get_worker_route(jk_map_t *m, const char *wname, const char *def);
 
 const char *jk_get_worker_domain(jk_map_t *m, const char *wname, const char *def);
@@ -123,6 +125,10 @@ const char *jk_get_lb_session_cookie(jk_map_t *m, const char *wname, const char
 
 const char *jk_get_lb_session_path(jk_map_t *m, const char *wname, const char *def);
 
+int jk_get_lb_set_session_cookie(jk_map_t *m, const char *wname, int def);
+
+const char *jk_get_lb_session_cookie_path(jk_map_t *m, const char *wname, const char *def);
+
 int jk_get_worker_retries(jk_map_t *m, const char *wname, int def);
 
 int jk_get_is_worker_disabled(jk_map_t *m, const char *wname);
@@ -181,6 +187,8 @@ int jk_is_unique_property(const char *prp_name);
 
 int jk_is_deprecated_property(const char *prp_name);
 
+int jk_check_buffer_size();
+    
 int jk_is_valid_property(const char *prp_name);
 
 int jk_get_worker_stdout(jk_map_t *m, const char *wname, const char **stdout_name);
@@ -229,7 +237,7 @@ const char *jk_get_worker_xml_doctype(jk_map_t *m, const char *wname, const char
 const char *jk_get_worker_prop_prefix(jk_map_t *m, const char *wname, const char *def);
 
 int jk_get_worker_fail_on_status(jk_map_t *m, const char *wname,
-                                 int *list, unsigned int list_size);
+                                 int **list, unsigned int *list_size);
 
 int jk_get_worker_user_case_insensitive(jk_map_t *m, const char *wname);
 
@@ -238,6 +246,8 @@ int is_http_status_fail(unsigned int http_status_fail_num,
 
 int jk_wildchar_match(const char *str, const char *exp, int icase);
 
+void jk_no2slash(char *name);
+
 #define TC32_BRIDGE_TYPE    32
 #define TC33_BRIDGE_TYPE    33
 #define TC40_BRIDGE_TYPE    40
